# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python implementations of preprocessing techniques.
# Contributors:
# ============================================================================

"""
Module of removal methods in the preprocessing stage:
- Many methods for removing stripe artifact in a sinogram (<-> ring artifact
in a reconstructed image).
- A zinger removal method.
- Blob removal methods.
"""

import numpy as np
import scipy.ndimage as ndi
from scipy import interpolate
import numpy.fft as fft
import algotom.util.utility as util


def remove_stripe_based_sorting(sinogram, size=21, dim=1, **options):
    """
    Remove stripe artifacts in a sinogram using the sorting technique,
    algorithm 3 in Ref. [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    size : int
        Window size of the median filter.
    dim : {1, 2}, optional
        Dimension of the window.
    options : dict, optional
        Use another smoothing filter rather than the median filter.
        E.g. options={"method": "gaussian_filter", "para1": (1,21))}

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    msg = "\n Please use the dictionary format: options={'method':" \
          " 'filter_name', 'para1': parameter_1, 'para2': parameter_2}"
    sino_sort, sino_index = util.sort_forward(np.float32(sinogram), axis=0)
    if len(options) == 0:
        if dim == 2:
            sino_sort = ndi.median_filter(sino_sort, (size, size))
        else:
            sino_sort = ndi.median_filter(sino_sort, (1, size))
    else:
        if not isinstance(options, dict):
            raise ValueError(msg)
        for opt_name in options:
            opt = options[opt_name]
            method = tuple(opt.values())[0]
            para = tuple(opt.values())[1:]
            if method in dir(ndi):
                try:
                    sino_sort = getattr(ndi, method)(sino_sort, *para)
                except:
                    raise ValueError(msg)
            else:
                if method in dir(util):
                    try:
                        sino_sort = getattr(util, method)(sino_sort, *para)
                    except:
                        raise ValueError(msg)
                else:
                    raise ValueError("Can't find the method: '{}' in the"
                                     " namespace".format(method))
    return util.sort_backward(sino_sort, sino_index, axis=0)


def remove_stripe_based_filtering(sinogram, sigma=3, size=21, dim=1, sort=True,
                                  **options):
    """
    Remove stripe artifacts in a sinogram using the filtering technique,
    algorithm 2 in Ref. [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    sigma : int
        Sigma of the Gaussian window used to separate the low-pass and
        high-pass components of the intensity profile of each column.
    size : int
        Window size of the median filter.
    dim : {1, 2}, optional
        Dimension of the window.
    sort : bool, optional
        Apply sorting if True.
    options : dict, optional
        Use another smoothing filter rather than the median filter.
        E.g. options={"method": "gaussian_filter", "para1": (1,21))}.

    Returns
    -------
    array_like
         2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    msg = "\n Please use the dictionary format: options={'method':" \
          " 'filter_name', 'para1': parameter_1, 'para2': parameter_2}"
    window = {"name": "gaussian", "sigma": sigma}
    sino_smooth, sino_sharp = util.separate_frequency_component(
        np.float32(sinogram), axis=0, window=window)
    if sort is True:
        sino_smooth, sino_index = util.sort_forward(sino_smooth, axis=0)
    if len(options) == 0:
        if dim == 2:
            sino_smooth = ndi.median_filter(sino_smooth, (size, size))
        else:
            sino_smooth = ndi.median_filter(sino_smooth, (1, size))
    else:
        if not isinstance(options, dict):
            raise ValueError(msg)
        for opt_name in options:
            opt = options[opt_name]
            method = tuple(opt.values())[0]
            if method in dir(ndi):
                para = tuple(opt.values())[1:]
                try:
                    sino_smooth = getattr(ndi, method)(sino_smooth, *para)
                except:
                    raise ValueError(msg)
            else:
                if method in dir(util):
                    try:
                        sino_smooth = getattr(util, method)(sino_smooth, *para)
                    except:
                        raise ValueError(msg)
                else:
                    raise ValueError("Can't find the method: '{}' in the"
                                     " namespace".format(method))
    if sort is True:
        sino_smooth = util.sort_backward(sino_smooth, sino_index, axis=0)
    return sino_smooth + sino_sharp


def remove_stripe_based_fitting(sinogram, order=2, sigma=10, sort=False,
                                num_chunk=1, **options):
    """
    Remove stripe artifacts in a sinogram using the fitting technique,
    algorithm 1 in Ref. [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    order : int
        Polynomial fit order.
    sigma : int
        Sigma of the Gaussian window in the x-direction. Smaller is stronger.
    sort : bool, optional
        Apply sorting if True.
    num_chunk : int
        Number of chunks of rows to apply the fitting.
    options : dict, optional
        Use another smoothing filter rather than the Fourier gaussian filter.
        E.g. options={"method": "gaussian_filter", "para1": (1,21))}.

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    msg = "\n Please use the dictionary format: options={'method':" \
          " 'filter_name', 'para1': parameter_1, 'para2': parameter_2}"
    (nrow, ncol) = sinogram.shape
    pad = min(150, int(0.1 * nrow))
    sigmay = np.clip(min(60, int(0.1 * ncol)), 10, None)
    if sort is True:
        sinogram, sino_index = util.sort_forward(sinogram, axis=0)
    sino_fit = util.generate_fitted_image(sinogram, order, axis=0,
                                          num_chunk=num_chunk)
    if len(options) == 0:
        sino_filt = util.apply_gaussian_filter(sino_fit, sigma, sigmay, pad)
    else:
        if not isinstance(options, dict):
            raise ValueError(msg)
        sino_filt = np.copy(sino_fit)
        for opt_name in options:
            opt = options[opt_name]
            method = tuple(opt.values())[0]
            if method in dir(ndi):
                para = tuple(opt.values())[1:]
                try:
                    sino_filt = getattr(ndi, method)(sino_filt, *para)
                except:
                    raise ValueError(msg)
            else:
                if method in dir(util):
                    try:
                        sino_filt = getattr(util, method)(sino_filt, *para)
                    except:
                        raise ValueError(msg)
                else:
                    raise ValueError("Can't find the method: '{}' in the"
                                     " namespace".format(method))
    sino_filt = np.mean(np.abs(sino_fit)) * sino_filt / np.mean(
        np.abs(sino_filt))
    sino_corr = ((sinogram / sino_fit) * sino_filt)
    if sort is True:
        sino_corr = util.sort_backward(sino_corr, sino_index, axis=0)
    return sino_corr


def remove_large_stripe(sinogram, snr=3.0, size=51, drop_ratio=0.1, norm=True,
                        **options):
    """
    Remove large stripe artifacts in a sinogram, algorithm 5 in Ref. [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    snr : float
        Ratio (>1.0) used to detect stripe locations. Greater is less sensitive.
    size : int
        Window size of the median filter.
    drop_ratio : float, optional
        Ratio of pixels to be dropped, which is used to to reduce
        the possibility of the false detection of stripes.
    norm : bool, optional
        Apply normalization if True.
    options : dict, optional
        Use another smoothing filter rather than the median filter.
        E.g. options={"method": "gaussian_filter", "para1": (1,21))}.

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    msg = "\n Please use the dictionary format: options={'method':" \
          " 'filter_name', 'para1': parameter_1, 'para2': parameter_2}"
    sinogram = np.copy(np.float32(sinogram))
    drop_ratio = np.clip(drop_ratio, 0.0, 0.8)
    (nrow, ncol) = sinogram.shape
    ndrop = int(0.5 * drop_ratio * nrow)
    sino_sort, sino_index = util.sort_forward(sinogram, axis=0)
    if len(options) == 0:
        sino_smooth = ndi.median_filter(sino_sort, (1, size))
    else:
        if not isinstance(options, dict):
            raise ValueError(msg)
        sino_smooth = np.copy(sino_sort)
        for opt_name in options:
            opt = options[opt_name]
            method = tuple(opt.values())[0]
            if method in dir(ndi):
                para = tuple(opt.values())[1:]
                try:
                    sino_smooth = getattr(ndi, method)(sino_smooth, *para)
                except:
                    raise ValueError(msg)
            else:
                if method in dir(util):
                    try:
                        sino_smooth = getattr(util, method)(sino_smooth, *para)
                    except:
                        raise ValueError(msg)
                else:
                    raise ValueError("Can't find the method: '{}' in the"
                                     " namespace".format(method))
    list1 = np.mean(sino_sort[ndrop:nrow - ndrop], axis=0)
    list2 = np.mean(sino_smooth[ndrop:nrow - ndrop], axis=0)
    list_fact = np.divide(list1, list2,
                          out=np.ones_like(list1), where=list2 != 0)
    list_mask = util.detect_stripe(list_fact, snr)
    list_mask = np.float32(ndi.binary_dilation(list_mask, iterations=1))
    if norm is True:
        sinogram = sinogram / np.tile(list_fact, (nrow, 1))
    sino_corr = util.sort_backward(sino_smooth, sino_index, axis=0)
    xlist_miss = np.where(list_mask > 0.0)[0]
    sinogram[:, xlist_miss] = sino_corr[:, xlist_miss]
    return sinogram


def remove_dead_stripe(sinogram, snr=3.0, size=51, residual=True):
    """
    Remove unresponsive or fluctuating stripe artifacts in a sinogram,
    algorithm 6 in Ref. [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    snr : float
        Ratio (>1.0) used to detect stripe locations. Greater is less sensitive.
    size : int
        Window size of the median filter.
    residual : bool, optional
        Removing residual stripes if True.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    sinogram = np.copy(sinogram)  # Make it mutable
    (nrow, _) = sinogram.shape
    sino_smooth = np.apply_along_axis(ndi.uniform_filter1d, 0, sinogram, 10)
    list_diff = np.sum(np.abs(sinogram - sino_smooth), axis=0)
    list_diff_bck = ndi.median_filter(list_diff, size)
    nmean = np.mean(np.abs(list_diff_bck))
    list_diff_bck[list_diff_bck == 0.0] = nmean
    list_fact = list_diff / list_diff_bck
    list_mask = util.detect_stripe(list_fact, snr)
    list_mask = np.float32(ndi.binary_dilation(list_mask, iterations=1))
    list_mask[0:2] = 0.0
    list_mask[-2:] = 0.0
    xlist = np.where(list_mask < 1.0)[0]
    ylist = np.arange(nrow)
    mat = sinogram[:, xlist]
    finter = interpolate.interp2d(xlist, ylist, mat, kind='linear')
    xlist_miss = np.where(list_mask > 0.0)[0]
    if len(xlist_miss) > 0:
        sinogram[:, xlist_miss] = finter(xlist_miss, ylist)
    if residual is True:
        sinogram = remove_large_stripe(sinogram, snr, size)
    return sinogram


def remove_all_stripe(sinogram, snr=3.0, la_size=51, sm_size=21, drop_ratio=0.1,
                      dim=1, **options):
    """
    Remove all types of stripe artifacts in a sinogram by combining algorithm
    6, 5, 4, and 3 in Ref. [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    snr : float
        Ratio (>1.0) used to detect stripe locations. Greater is less sensitive.
    la_size : int
        Window size of the median filter to remove large stripes.
    sm_size : int
        Window size of the median filter to remove small-to-medium stripes.
    drop_ratio : float, optional
        Ratio of pixels to be dropped, which is used to to reduce
        the possibility of the false detection of stripes.
    dim : {1, 2}, optional
        Dimension of the window.
    options : dict, optional
        Use another smoothing filter rather than the median filter.
        E.g. options={"method": "gaussian_filter", "para1": (1,21))}

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    sinogram = remove_dead_stripe(sinogram, snr, la_size, residual=False)
    sinogram = remove_large_stripe(sinogram, snr, la_size, drop_ratio,
                                   **options)
    sinogram = remove_stripe_based_sorting(sinogram, sm_size, dim, **options)
    return sinogram


def remove_stripe_based_2d_filtering_sorting(sinogram, sigma=3, size=21, dim=1,
                                             **options):
    """
    Remove stripes using a 2D low-pass filter and the sorting-based technique,
    algorithm in section 3.3.4 in Ref. [1].
    Angular direction is along the axis 0.

    Parameters
    ---------
    sinogram : array_like
        2D array. Sinogram image.
    sigma : int
        Sigma of the Gaussian window.
    size : int
        Window size of the median filter.
    dim : {1, 2}, optional
        Dimension of the window.

    Returns
    -------
    array_like
         2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1117/12.2530324
    """
    (nrow, ncol) = sinogram.shape
    pad = min(150, int(0.1 * min(nrow, ncol)))
    sino_smooth = util.apply_gaussian_filter(sinogram, sigma, sigma, pad)
    sino_sharp = sinogram - sino_smooth
    sino_sharp = remove_stripe_based_sorting(sino_sharp, size, dim, **options)
    return sino_smooth + sino_sharp


def remove_stripe_based_normalization(sinogram, sigma=15, num_chunk=1,
                                      sort=True, **options):
    """
    Remove stripes using the method in Ref. [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    sigma : int
        Sigma of the Gaussian window.
    num_chunk : int
        Number of chunks of rows.
    sort : bool, optional
        Apply sorting (Ref. [2]) if True.
    options : dict, optional
        Use another smoothing 1D-filter rather than the Gaussian filter.
        E.g. options={"method": "median_filter", "para1": 21)}.

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://www.mcs.anl.gov/research/projects/X-ray-cmt/rivers/
           tutorial.html
    .. [2] https://doi.org/10.1364/OE.26.028396
    """
    msg = "\n Please use the dictionary format: options={'method':" \
          " 'filter_name', 'para1': parameter_1, 'para2': parameter_2}" \
          "\n Note that the filter must be a 1D-filter."
    (nrow, _) = sinogram.shape
    sinogram = np.copy(sinogram)
    if sort is True:
        sinogram, sino_index = util.sort_forward(sinogram, axis=0)
    list_index = np.array_split(np.arange(nrow), num_chunk)
    for pos in list_index:
        bindex = pos[0]
        eindex = pos[-1] + 1
        list_mean = np.mean(sinogram[bindex:eindex], axis=0)
        if len(options) == 0:
            list_filt = ndi.gaussian_filter(list_mean, sigma)
        else:
            if not isinstance(options, dict):
                raise ValueError(msg)
            list_filt = np.copy(list_mean)
            for opt_name in options:
                opt = options[opt_name]
                method = tuple(opt.values())[0]
                if method in dir(ndi):
                    para = tuple(opt.values())[1:]
                    try:
                        list_filt = getattr(ndi, method)(list_filt, *para)
                    except:
                        raise ValueError(msg)
                else:
                    if method in dir(util):
                        try:
                            list_filt = getattr(util, method)(list_filt, *para)
                        except:
                            raise ValueError(msg)
                    else:
                        raise ValueError("Can't find the method: '{}' in the"
                                         " namespace".format(method))
        list_coe = list_filt - list_mean
        matcoe = np.tile(list_coe, (eindex - bindex, 1))
        sinogram[bindex:eindex, :] = sinogram[bindex:eindex, :] + matcoe
    if sort is True:
        sinogram = util.sort_backward(sinogram, sino_index, axis=0)
    return sinogram


def remove_stripe_based_regularization(sinogram, alpha=0.0005, num_chunk=1,
                                       apply_log=True, sort=True):
    """
    Remove stripes using the method in Ref. [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    alpha : float
        Regularization parameter, e.g. 0.0005. Smaller is stronger.
    num_chunk : int
        Number of chunks of rows.
    apply_log : bool
        Apply the logarithm function to the sinogram if True.
    sort : bool, optional
        Apply sorting (Ref. [2]) if True.

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1016/j.aml.2010.08.022
    .. [2] https://doi.org/10.1364/OE.26.028396
    """
    (nrow, ncol) = sinogram.shape
    sinogram = np.copy(sinogram)
    if sort is True:
        sinogram, sino_index = util.sort_forward(sinogram, axis=0)
    if apply_log is True:
        if np.any(sinogram <= 0.0):
            nmean = np.mean(np.abs(sinogram))
            sinogram[sinogram <= 0.0] = nmean
            sinogram = -np.log(sinogram)
        else:
            sinogram = -np.log(sinogram)
    sijmat = util.calculate_regularization_coefficient(ncol, alpha)
    list_index = np.array_split(np.arange(nrow), num_chunk)
    list_grad = np.zeros(ncol, dtype=np.float32)
    mat_grad = np.zeros((ncol, ncol), dtype=np.float32)
    for pos in list_index:
        bindex = pos[0]
        eindex = pos[-1] + 1
        list_mean = np.mean(sinogram[bindex:eindex], axis=0)
        list_grad[1:-1] = (-1) * np.diff(list_mean, 2)
        list_grad[0] = list_mean[0] - list_mean[1]
        list_grad[-1] = list_mean[-1] - list_mean[-2]
        mat_grad[:] = list_grad
        list_coe = np.sum(mat_grad * sijmat, axis=1)
        mat_coe = np.tile(list_coe, (eindex - bindex, 1))
        sinogram[bindex:eindex, :] = sinogram[bindex:eindex, :] + mat_coe
    if sort is True:
        sinogram = util.sort_backward(sinogram, sino_index, axis=0)
    if apply_log is True:
        sinogram = np.exp(-sinogram)
    return sinogram


def remove_stripe_based_fft(sinogram, u=20, n=8, v=1, sort=False):
    """
    Remove stripes using the method in Ref. [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    u : int
        Cutoff frequency.
    n : int
        Filter order.
    v : int
        Number of rows (* 2) to be applied the filter.
    sort : bool, optional
        Apply sorting (Ref. [2]) if True.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1063/1.1149043
    .. [2] https://doi.org/10.1364/OE.26.028396
    """
    if sort is True:
        sinogram, sino_index = util.sort_forward(sinogram, axis=0)
    pad = min(150, int(0.1 * np.min(sinogram.shape)))
    sinogram = np.pad(sinogram, ((pad, pad), (0, 0)), mode='mean')
    sinogram = np.pad(sinogram, ((0, 0), (pad, pad)), mode='edge')
    (nrow, ncol) = sinogram.shape
    window_2d = util.make_2d_butterworth_window(ncol, nrow, u, v, n)
    sinogram = fft.ifft2(
        np.fft.ifftshift(np.fft.fftshift(fft.fft2(sinogram)) * window_2d))
    sinogram = np.abs(sinogram[pad:nrow - pad, pad:ncol - pad])
    if sort is True:
        sinogram = util.sort_backward(sinogram, sino_index, axis=0)
    return sinogram


def remove_stripe_based_wavelet_fft(sinogram, level=5, size=1,
                                    wavelet_name="db9", window_name="gaussian",
                                    sort=False, **options):
    """
    Remove stripes using the method in Ref. [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    level : int
        Wavelet decomposition level.
    size : int
        Damping parameter. Larger is stronger.
    wavelet_name : str
        Name of a wavelet. Search pywavelets API for a full list.
    window_name : str
        High-pass window. Two options: "gaussian" or "butter".
    sort : bool, optional
        Apply sorting (Ref. [2]) if True.

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.17.008567
    .. [2] https://doi.org/10.1364/OE.26.028396
    """
    msg = "\n Please use the dictionary format: options={'method':" \
          " 'filter_name', 'para1': parameter_1, 'para2': parameter_2}"
    if sort is True:
        sinogram, sino_index = util.sort_forward(sinogram, axis=0)
    (nrow, ncol) = sinogram.shape
    pad = min(150, int(0.1 * min(nrow, ncol)))
    sinogram = np.pad(sinogram, ((pad, pad), (0, 0)), mode='mean')
    sinogram = np.pad(sinogram, ((0, 0), (pad, pad)), mode='edge')
    output_data = util.apply_wavelet_decomposition(sinogram, wavelet_name,
                                                   level=level)
    output_data = [list(data) for data in output_data]
    n_level = len(output_data[1:])
    for i in range(1, n_level + 1):
        if len(options) == 0:
            (height, width) = output_data[i][1].shape
            window = np.transpose(
                util.make_2d_damping_window(height, width, size, window_name))
            output_data[i][1] = np.real(np.fft.ifft2(np.fft.ifftshift(
                np.fft.fftshift(np.fft.fft2(output_data[i][1])) * window)))
        else:
            if not isinstance(options, dict):
                raise ValueError(msg)
            mat_smooth = np.copy(output_data[i][1])
            for opt_name in options:
                opt = options[opt_name]
                method = tuple(opt.values())[0]
                para = tuple(opt.values())[1:]
                if method in dir(ndi):
                    try:
                        mat_smooth = getattr(ndi, method)(mat_smooth, *para)
                    except:
                        raise ValueError(msg)
                elif method in dir(util):
                    try:
                        mat_smooth = getattr(util, method)(mat_smooth, *para)
                    except:
                        raise ValueError(msg)
                else:
                    raise ValueError("Can't find the method: '{}' in the"
                                     " namespace".format(method))
            output_data[i][1] = mat_smooth
    sinogram = util.apply_wavelet_recontruction(output_data, wavelet_name)
    sinogram = sinogram[pad:nrow + pad, pad:ncol + pad]
    if sort is True:
        sinogram = util.sort_backward(sinogram, sino_index, axis=0)
    return sinogram


def remove_stripe_based_interpolation(sinogram, snr=3.0, size=51,
                                      drop_ratio=0.1, norm=True, kind="linear",
                                      **options):
    """
    Combination of algorithm 4, 5, and 6 in Ref. [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    snr : float
        Ratio (>1.0) used to detect stripe locations. Greater is less sensitive.
    size : int
        Window size of the median filter used to detect stripes.
    drop_ratio : float, optional
        Ratio of pixels to be dropped, which is used to to reduce
        the possibility of the false detection of stripes.
    norm : bool, optional
        Apply normalization if True.
    kind : {'linear', 'cubic', 'quintic'}, optional
        The kind of spline interpolation to use. Default is 'linear'.
    options : dict, optional
        Use another smoothing filter rather than the median filter.
        E.g. options={"method": "gaussian_filter", "para1": (1,21))}

    Returns
    -------
    array_like
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    msg = "\n Please use the dictionary format: options={'method':" \
          " 'filter_name', 'para1': parameter_1, 'para2': parameter_2}"
    drop_ratio = np.clip(drop_ratio, 0.0, 0.8)
    sinogram = np.copy(sinogram)
    (nrow, ncol) = sinogram.shape
    ndrop = int(0.5 * drop_ratio * nrow)
    sino_sort = np.sort(sinogram, axis=0)
    if len(options) == 0:
        sino_smooth = ndi.median_filter(sino_sort, (1, size))
    else:
        if not isinstance(options, dict):
            raise ValueError(msg)
        sino_smooth = np.copy(sino_sort)
        for opt_name in options:
            opt = options[opt_name]
            method = tuple(opt.values())[0]
            para = tuple(opt.values())[1:]
            if method in dir(ndi):
                try:
                    sino_smooth = getattr(ndi, method)(sino_smooth, *para)
                except:
                    raise ValueError(msg)
            else:
                if method in dir(util):
                    try:
                        sino_smooth = getattr(util, method)(sino_smooth, *para)
                    except:
                        raise ValueError(msg)
                else:
                    raise ValueError("Can't find the method: '{}' in the"
                                     " namespace".format(method))
    list1 = np.mean(sino_sort[ndrop:nrow - ndrop], axis=0)
    list2 = np.mean(sino_smooth[ndrop:nrow - ndrop], axis=0)
    list_fact = np.divide(list1, list2,
                          out=np.ones_like(list1), where=list2 != 0)
    list_mask = util.detect_stripe(list_fact, snr)
    list_mask = np.float32(ndi.binary_dilation(list_mask, iterations=1))
    mat_fact = np.tile(list_fact, (nrow, 1))
    if norm is True:
        sinogram = sinogram / mat_fact
    list_mask[0:2] = 0.0
    list_mask[-2:] = 0.0
    xlist = np.where(list_mask < 1.0)[0]
    ylist = np.arange(nrow)
    zmat = sinogram[:, xlist]
    finter = interpolate.interp2d(xlist, ylist, zmat, kind=kind)
    xlist_miss = np.where(list_mask > 0.0)[0]
    if len(xlist_miss) > 0:
        sinogram[:, xlist_miss] = finter(xlist_miss, ylist)
    return sinogram


def check_zinger_size(mat, max_size):
    """
    Check if the size of a zinger is smaller than a given size.

    Parameters
    ----------
    mat : array_like
        2D array.
    max_size : int
        Maximum size.

    Returns
    -------
    bool
    """
    check = False
    zinger_size = mat.sum()
    if zinger_size <= max_size:
        check = True
    return check


def select_zinger(mat, max_size):
    """
    Select zingers smaller than a certain size.

    Parameters
    ----------
    mat : array_like
        2D array.
    max_size : int
        Maximum size in pixel.

    Returns
    -------
    array_like
        2D binary array.
    """
    list_zin = ndi.find_objects(ndi.label(mat)[0])
    zin_sel = [zin for zin in list_zin
               if check_zinger_size(mat[zin], max_size)]
    mat_out = np.zeros_like(mat)
    for _, j in enumerate(zin_sel):
        mat_out[j] = mat[j]
    return mat_out


def remove_zinger(mat, threshold, size=2):
    """
    Remove zinger using the method in Ref. [1], working on a projection image
    or sinogram image.

    Parameters
    ----------
    mat : array_like
        2D array. Projection image or sinogram image.
    threshold : float
        Threshold to segment zingers. Smaller is more sensitive.
        Recommended range [0.05, 0.1].
    size : int
        Size of a zinger.

    Returns
    -------
    array_like
        2D array. Zinger-removed image.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448

    """
    step = np.clip(size, 1, None)
    mat = np.copy(mat)
    mat_ave = []
    for i in range(-size, size + 1, step):
        for j in range(-size, size + 1, step):
            if (i != 0) or (j != 0):
                mat_ave.append(np.roll(np.roll(mat, i, axis=0), j, axis=1))
    mat_ave = np.mean(np.asarray(mat_ave), axis=0)
    mat_ave[mat_ave == 0.0] = 1.0
    mat_nor = mat / mat_ave - 1.0
    mask = np.asarray(mat_nor > threshold, dtype=np.float32)
    mask = select_zinger(mask, size)
    mat[mask > 0.0] = mat_ave[mask > 0.0]
    return mat


def generate_blob_mask(flat, size, snr):
    """
    Generate a binary mask of blobs from a flat-field image (Ref. [1]).

    Parameters
    ----------
    flat : array_like
        2D array. Flat-field image.
    size : float
        Estimated size of the largest blob.
    snr : float
        Ratio used to segment blobs.

    Returns
    -------
    array_like
        2D array. Binary mask.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    mat = ndi.median_filter(flat, (2, 2))
    mask = np.zeros_like(mat)
    for i in range(mat.shape[0]):
        line = mat[i]
        line_fil = ndi.median_filter(line, size)
        line_fil[line_fil == 0.0] = np.mean(line_fil)
        line_norm = line / line_fil
        mask_1d = util.detect_stripe(line_norm, snr)
        mask_1d[0:2] = 0.0
        mask_1d[-2:] = 0.0
        mask[i] = np.float32(ndi.binary_dilation(mask_1d, iterations=1))
    return mask


def remove_blob_1d(sino_1d, mask_1d):
    """
    Remove blobs in one row of a sinogram, e.g. for a helical scan as shown in
    Ref. [1].

    Parameters
    ----------
    sino_1d : array_like
        1D array. A row of a sinogram.
    mask_1d : array_like
        1D binary mask.

    Returns
    -------
    array_like
        1D array.

    Notes
    -----
    The method is used to remove streak artifacts caused by blobs in
    a sinogram generated from a helical-scan data [1].

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    sino_1d = np.copy(sino_1d)
    listx = np.where(mask_1d < 1.0)[0]
    listy = sino_1d[listx]
    finter = interpolate.interp1d(listx, listy)
    mask_1d[:2] = 0.0
    mask_1d[-2:] = 0.0
    listx_miss = np.where(mask_1d > 0.0)[0]
    if len(listx_miss) > 0:
        sino_1d[listx_miss] = finter(listx_miss)
    return sino_1d


def remove_blob(mat, mask):
    """
    Remove blobs in an image.

    Parameters
    ----------
    mat : array_like
        2D array. Projection image or sinogram image.
    mask : array_like
        2D binary mask.

    Returns
    -------
    array_like
        2D array.
    """
    mat = np.copy(mat)
    if mat.shape != mask.shape:
        raise ValueError("The image and the mask are not the same shape !!!")
    for i in range(mat.shape[0]):
        array_1d = mat[i]
        mask_1d = mask[i]
        mask_1d[:2] = 0.0
        mask_1d[-2:] = 0.0
        listx = np.where(mask_1d < 1.0)[0]
        listy = array_1d[listx]
        finter = interpolate.interp1d(listx, listy)
        listx_miss = np.where(mask_1d > 0.0)[0]
        if len(listx_miss) > 0:
            array_1d[listx_miss] = finter(listx_miss)
        mat[i] = array_1d
    return mat



# ===========================================================================
# ===========================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===========================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python codes for converting data format.
# Contributors:
# ===========================================================================

"""
Module for converting data type:
- Convert a list of tif files to a hdf/nxs file.
- Extract tif images from a hdf/nxs file.
"""

import os
import numpy as np
import algotom.io.loadersaver as losa


def convert_tif_to_hdf(input_path, output_path, key_path="entry/data",
                       crop=(0, 0, 0, 0), pattern=None, **options):
    """
    Convert a folder of tif files to a hdf/nxs file.

    Parameters
    ----------
    input_path : str
        Folder path to the tif files.
    output_path : str
        Path to the hdf/nxs file.
    key_path : str, optional
        Key path to the dataset.
    crop : tuple of int, optional
        Crop the images from the edges, i.e.
        crop = (crop_top, crop_bottom, crop_left, crop_right).
    pattern : str, optional
        Used to find tif files with names matching the pattern.
    options : dict, optional
        Add metadata. E.g. options={"entry/angles": angles, "entry/energy": 53}.

    Returns
    -------
    str
        Path to the hdf/nxs file.
    """
    if pattern is None:
        list_file = losa.find_file(input_path + "/*.tif*")
    else:
        list_file = losa.find_file(input_path + "/*" + pattern + "*.tif*")
    depth = len(list_file)
    if depth == 0:
        raise ValueError("No tif files in the folder: {}".format(input_path))
    (height, width) = np.shape(losa.load_image(list_file[0]))
    file_base, file_ext = os.path.splitext(output_path)
    if not (file_ext == '.hdf' or file_ext == '.h5' or file_ext == ".nxs"):
        file_ext = '.hdf'
    output_path = file_base + file_ext
    cr_top, cr_bottom, cr_left, cr_right = crop
    cr_height = height - cr_top - cr_bottom
    cr_width = width - cr_left - cr_right
    data_out = losa.open_hdf_stream(output_path, (depth, cr_height, cr_width),
                                    key_path=key_path, overwrite=True,
                                    **options)
    for i, fname in enumerate(list_file):
        data_out[i] = losa.load_image(fname)[cr_top:cr_height + cr_top,
                      cr_left:cr_width + cr_left]
    return output_path


def extract_tif_from_hdf(input_path, output_path, key_path, index=(0, -1, 1),
                         axis=0, crop=(0, 0, 0, 0), prefix="img"):
    """
    Extract tif images from a hdf/nxs file.

    Parameters
    ----------
    input_path : str
        Path to the hdf/nxs file.
    output_path : str
        Output folder.
    key_path : str
        Key path to the dataset in the hdf/nxs file.
    index : tuple of int or int.
        Indices of extracted images. A tuple corresponds to (start,stop,step).
    axis : int
        Axis which the images are extracted.
    crop : tuple of int, optional
        Crop the images from the edges, i.e.
        crop = (crop_top, crop_bottom, crop_left, crop_right).
    prefix : str, optional
        Prefix of names of tif files.

    Returns
    -------
    str
        Folder path to the tif files.
    """
    data = losa.load_hdf(input_path, key_path)
    (depth, height, width) = data.shape
    if isinstance(index, tuple):
        start, stop, step = index
    else:
        start, stop, step = index, index + 1, 1
    cr_top, cr_bottom, cr_left, cr_right = crop
    if axis == 1:
        if (stop == -1) or (stop > height):
            stop = height
        for i in range(start, stop, step):
            mat = data[cr_top:depth - cr_bottom, i, cr_left:width - cr_right]
            out_name = "0000" + str(i)
            losa.save_image(
                output_path + "/" + prefix + "_" + out_name[-5:] + ".tif", mat)
    elif axis == 2:
        if (stop == -1) or (stop > width):
            stop = width
        for i in range(start, stop, step):
            mat = data[cr_top:depth - cr_bottom, cr_left:height - cr_right, i]
            out_name = "0000" + str(i)
            losa.save_image(
                output_path + "/" + prefix + "_" + out_name[-5:] + ".tif", mat)
    else:
        if (stop == -1) or (stop > depth):
            stop = depth
        for i in range(start, stop, step):
            mat = data[i, cr_top:height - cr_bottom, cr_left:width - cr_right]
            out_name = "0000" + str(i)
            losa.save_image(
                output_path + "/" + prefix + "_" + out_name[-5:] + ".tif", mat)
    return output_path



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python implementations of preprocessing techniques.
# Contributors:
# ============================================================================

"""
Module of conversion methods in the preprocessing stage:
- Stitching images.
- Joining images if there is no overlapping.
- Converting a 360-degree sinogram with offset center-of-rotation (COR) to
  a 180-degree sinogram.
- Extending a 360-degree sinogram with offset COR for direct reconstruction
  instead of converting it to a 180-degree sinogram.
- Converting a 180-degree sinogram to a 360-sinogram.
- Generating a sinogram from a helical data.
"""

import numpy as np
from scipy import interpolate
from scipy.ndimage import shift
import algotom.prep.removal as remo
import algotom.prep.calculation as calc


def make_weight_matrix(mat1, mat2, overlap, side):
    """
    Generate a linear-ramp weighting matrix for image stitching.

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image or sinogram image.
    mat2 :  array_like
        2D array. Projection image or sinogram image.
    overlap : int
        Width of the overlap area between two images.
    side : {0, 1}
        Only two options: 0 or 1. It is used to indicate the overlap side
        respects to image 1. "0" corresponds to the left side. "1" corresponds
        to the right side.
    """
    overlap = int(np.floor(overlap))
    wei_mat1 = np.ones_like(mat1)
    wei_mat2 = np.ones_like(mat2)
    if side == 1:
        list_down = np.linspace(1.0, 0.0, overlap)
        list_up = 1.0 - list_down
        wei_mat1[:, -overlap:] = np.float32(list_down)
        wei_mat2[:, :overlap] = np.float32(list_up)
    else:
        list_down = np.linspace(1.0, 0.0, overlap)
        list_up = 1.0 - list_down
        wei_mat2[:, -overlap:] = np.float32(list_down)
        wei_mat1[:, :overlap] = np.float32(list_up)
    return wei_mat1, wei_mat2


def stitch_image(mat1, mat2, overlap, side, wei_mat1=None, wei_mat2=None,
                 norm=True, total_width=None):
    """
    Stitch projection images or sinogram images using a linear ramp.

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image or sinogram image.
    mat2 :  array_like
        2D array. Projection image or sinogram image.
    overlap : float
        Width of the overlap area between two images.
    side : {0, 1}
        Only two options: 0 or 1. It is used to indicate the overlap side
        respects to image 1. "0" corresponds to the left side. "1" corresponds
        to the right side.
    wei_mat1 : array_like, optional
        Weighting matrix used for image 1.
    wei_mat2 : array_like, optional
        Weighting matrix used for image 2.
    norm : bool, optional
        Enable/disable normalization before stitching.
    total_width : int, optional
        Final width of the stitched image.

    Returns
    -------
    array_like
        Stitched image.
    """
    (nrow1, ncol1) = mat1.shape
    (nrow2, ncol2) = mat2.shape
    overlap_int = int(np.floor(overlap))
    sub_pixel = overlap - overlap_int
    if sub_pixel > 0.0:
        if side == 1:
            mat1 = shift(mat1, (0, sub_pixel), mode='nearest')
            mat2 = shift(mat2, (0, -sub_pixel), mode='nearest')
        else:
            mat1 = shift(mat1, (0, -sub_pixel), mode='nearest')
            mat2 = shift(mat2, (0, sub_pixel), mode='nearest')
    if nrow1 != nrow2:
        raise ValueError("Two images are not at the same height!!!")
    if (wei_mat1 is None) or (wei_mat2 is None):
        (wei_mat1, wei_mat2) = make_weight_matrix(mat1, mat2, overlap_int, side)
    total_width0 = ncol1 + ncol2 - overlap_int
    if (total_width is None) or (total_width < total_width0):
        total_width = total_width0
    mat_comb = np.zeros((nrow1, total_width0), dtype=np.float32)
    if side == 1:
        if norm is True:
            factor1 = np.mean(mat1[:, -overlap_int:])
            factor2 = np.mean(mat2[:, :overlap_int])
            mat2 = mat2 * factor1 / factor2
        mat_comb[:, 0:ncol1] = mat1 * wei_mat1
        mat_comb[:, (ncol1 - overlap_int):total_width0] += mat2 * wei_mat2
    else:
        if norm is True:
            factor2 = np.mean(mat2[:, -overlap_int:])
            factor1 = np.mean(mat1[:, :overlap_int])
            mat2 = mat2 * factor1 / factor2
        mat_comb[:, 0:ncol2] = mat2 * wei_mat2
        mat_comb[:, (ncol2 - overlap_int):total_width0] += mat1 * wei_mat1
    if total_width > total_width0:
        mat_comb = np.pad(
            mat_comb, ((0, 0), (0, total_width - total_width0)), mode='edge')
    return mat_comb


def join_image(mat1, mat2, joint_width, side, norm=True, total_width=None):
    """
    Join projection images or sinogram images. This is useful for fixing the
    problem of non-overlap between images.

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image or sinogram image.
    mat2 :  array_like
        2D array. Projection image or sinogram image.
    joint_width : float
        Width of the joint area between two images.
    side : {0, 1}
        Only two options: 0 or 1. It is used to indicate the overlap side
        respects to image 1. "0" corresponds to the left side. "1" corresponds
        to the right side.
    norm : bool
        Enable/disable normalization before joining.
    total_width : int, optional
        Final width of the joined image.

    Returns
    -------
    array_like
        Stitched image.
    """
    (nrow1, ncol1) = mat1.shape
    (nrow2, ncol2) = mat2.shape
    joint_int = int(np.floor(joint_width))
    sub_pixel = joint_width - joint_int
    side = int(side)
    if sub_pixel > 0.0:
        if side == 1:
            mat1 = shift(mat1, (0, sub_pixel), mode='nearest')
            mat2 = shift(mat2, (0, -sub_pixel), mode='nearest')
        else:
            mat1 = shift(mat1, (0, -sub_pixel), mode='nearest')
            mat2 = shift(mat2, (0, sub_pixel), mode='nearest')
    if nrow1 != nrow2:
        raise ValueError("Two images are not at the same height!!!")
    total_width0 = ncol1 + ncol2 + joint_int
    if (total_width is None) or (total_width < total_width0):
        total_width = total_width0
    mat_comb = np.zeros((nrow1, total_width0), dtype=np.float32)
    if side == 1:
        if norm is True:
            factor1 = np.mean(mat1[:, -3:])
            factor2 = np.mean(mat2[:, :3])
            mat2 = mat2 * factor1 / factor2
        mat_comb[:, 0:ncol1] = mat1
        mat_comb[:, (ncol1 + joint_int):total_width0] += mat2
        list_mask = np.zeros(total_width0, dtype=np.float32)
        list_mask[ncol1 - 2:ncol1 + joint_int + 3] = 1.0
        listx = np.where(list_mask < 1.0)[0]
        listy = np.arange(nrow1)
        mat = mat_comb[:, listx]
        finter = interpolate.interp2d(listx, listy, mat, kind='linear')
        listx_miss = np.where(list_mask > 0.0)[0]
        if len(listx_miss) > 0:
            mat_comb[:, listx_miss] = finter(listx_miss, listy)
    else:
        if norm is True:
            factor2 = np.mean(mat2[:, -3:])
            factor1 = np.mean(mat1[:, :3])
            mat2 = mat2 * factor1 / factor2
        mat_comb[:, 0:ncol2] = mat2
        mat_comb[:, (ncol2 + joint_int):total_width0] += mat1
        list_mask = np.zeros(total_width0, dtype=np.float32)
        list_mask[ncol2 - 2:ncol2 + joint_int + 3] = 1.0
        listx = np.where(list_mask < 1.0)[0]
        listy = np.arange(nrow1)
        mat = mat_comb[:, listx]
        finter = interpolate.interp2d(listx, listy, mat, kind='linear')
        listx_miss = np.where(list_mask > 0.0)[0]
        if len(listx_miss) > 0:
            mat_comb[:, listx_miss] = finter(listx_miss, listy)
    if total_width > total_width0:
        mat_comb = np.pad(
            mat_comb, ((0, 0), (0, total_width - total_width0)), mode='edge')
    return mat_comb


def stitch_image_multiple(list_mat, list_overlap, norm=True, total_width=None):
    """
    Stitch list of projection images or sinogram images using a linear ramp.

    Parameters
    ----------
    list_mat : list of array_like
        List of 2D array. Projection image or sinogram image.
    list_overlap : list of tuple of floats
        List of [overlap, side].
        overlap : Width of the overlap area between two images.
        side : Overlap side between two images.
    norm : bool, optional
        Enable/disable normalization before stitching.
    total_width : int, optional
        Final width of the stitched image.

    Returns
    -------
    array_like
        Stitched image.
    """
    num_mat = len(list_mat)
    mat_comb = np.copy(list_mat[0])
    if num_mat > 1:
        for i in range(1, num_mat):
            (overlap, side) = list_overlap[i - 1][0:2]
            mat_comb = stitch_image(mat_comb, list_mat[i], overlap, side, norm)
        width = mat_comb.shape[1]
        if total_width is None:
            total_width = width
        if total_width > width:
            mat_comb = np.pad(
                mat_comb, ((0, 0), (0, total_width - width)), mode='edge')
    else:
        raise ValueError("Need at least 2 images to work!!!")
    return np.asarray(mat_comb)


def join_image_multiple(list_mat, list_joint, norm=True, total_width=None):
    """
    Join list of projection images or sinogram images. This is useful for
    fixing the problem of non-overlap between images.

    Parameters
    ----------
    list_mat : list of array_like
        List of 2D array. Projection image or sinogram image.
    list_joint : list of tuple of floats
        List of [joint_width, side].
        joint_width : Width of the joint area between two images.
        side : Overlap side between two images.
    norm : bool, optional
        Enable/disable normalization before stitching.
    total_width : int, optional
        Final width of the stitched image.

    Returns
    -------
    array_like
        Stitched image.
    """
    num_mat = len(list_mat)
    if num_mat > 1:
        mat_comb = np.copy(list_mat[0])
        for i in range(1, num_mat):
            (joint_width, side) = list_joint[i - 1][0:2]
            mat_comb = join_image(mat_comb, list_mat[i], joint_width, side,
                                  norm)
        width = mat_comb.shape[1]
        if total_width is None:
            total_width = width
        if total_width > width:
            mat_comb = np.pad(
                mat_comb, ((0, 0), (0, total_width - width)), mode='edge')
    else:
        raise ValueError("Need at least 2 images to work!!!")
    return np.asarray(mat_comb)


def convert_sinogram_360_to_180(sino_360, cor, wei_mat1=None, wei_mat2=None,
                                norm=True, total_width=None):
    """
    Convert a 360-degree sinogram to a 180-degree sinogram.

    Parameters
    ----------
    sino_360 : array_like
        2D array. 360-degree sinogram.
    cor : float or tuple of float
        Center-of-rotation or (Overlap_area, overlap_side).
    wei_mat1 : array_like, optional
        Weighting matrix used for the 1st haft of the sinogram.
    wei_mat2 : array_like, optional
        Weighting matrix used for the 2nd haft of the sinogram.
    norm : bool, optional
        Enable/disable normalization before stitching.
    total_width : int, optional
        Final width of the stitched image.

    Returns
    -------
    sino_stiched : array_like
        Converted sinogram.
    cor : float
        Updated center-of-rotation referred to the converted sinogram.
    """
    (nrow, ncol) = sino_360.shape
    xcenter = (ncol - 1.0) * 0.5
    nrow_180 = nrow // 2 + 1
    sino_top = sino_360[0:nrow_180, :]
    sino_bot = np.fliplr(sino_360[-nrow_180:, :])
    if isinstance(cor, tuple):
        (overlap, side) = cor
    else:
        if cor <= xcenter:
            overlap = 2 * (cor + 1)
            side = 0
        else:
            overlap = 2 * (ncol - cor - 1)
            side = 1
    sino_stitch = stitch_image(
        sino_top, sino_bot, overlap, side, wei_mat1=wei_mat1,
        wei_mat2=wei_mat2, norm=norm, total_width=total_width)
    cor = (2 * ncol - np.floor(overlap) - 1.0) / 2.0
    return sino_stitch, cor


def convert_sinogram_180_to_360(sino_180, center):
    """
    Convert a 180-degree sinogram to a 360-degree sinogram (Ref. [1]).

    Parameters
    ----------
    sino_180 : array_like
        2D array. 180-degree sinogram.
    center : float
        Center-of-rotation.

    Returns
    -------
    array_like
        360-degree sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.22.019078
    """
    (nrow, ncol) = sino_180.shape
    xcenter = (ncol - 1.0) / 2.0
    shift_x = xcenter - center
    sino_flip = shift(np.fliplr(shift(sino_180, (0, shift_x), mode='nearest')),
                      (0, -shift_x), mode='nearest')
    return np.vstack((sino_180, sino_flip[1:]))


def extend_sinogram(sino_360, cor, apply_log=True):
    """
    Extend a 360-degree sinogram (with offset center-of-rotation) for
    later reconstruction (Ref. [1]).

    Parameters
    ----------
    sino_360 : array_like
        2D array. 360-degree sinogram.
    cor : float or tuple of float
        Center-of-rotation or (Overlap_area, overlap_side).
    apply_log : bool, optional
        Apply the logarithm function if True.

    Returns
    -------
    sino_pad : array_like
        Extended sinogram.
    cor : float
        Updated center-of-rotation referred to the converted sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    if apply_log is True:
        sino_360 = -np.log(sino_360)
    else:
        sino_360 = np.copy(sino_360)
    (nrow, ncol) = sino_360.shape
    xcenter = (ncol - 1.0) * 0.5
    if isinstance(cor, tuple):
        (overlap, side) = cor
    else:
        if cor <= xcenter:
            overlap = 2 * (cor + 1)
            side = 0
        else:
            overlap = 2 * (ncol - cor - 1)
            side = 1
    overlap_int = int(np.floor(overlap))
    sub_pixel = overlap - overlap_int
    if side == 1:
        if sub_pixel > 0.0:
            sino_360 = shift(sino_360, (0, sub_pixel), mode='nearest')
        wei_list = np.linspace(1.0, 0.0, overlap_int)
        wei_mat = np.tile(wei_list, (nrow, 1))
        sino_360[:, -overlap_int:] = sino_360[:, -overlap_int:] * wei_mat
        pad_wid = ncol - overlap_int
        sino_pad = np.pad(sino_360, ((0, 0), (0, pad_wid)), mode='edge')
    else:
        if sub_pixel > 0.0:
            sino_360 = shift(sino_360, (0, -sub_pixel), mode='nearest')
        wei_list = np.linspace(0.0, 1.0, overlap_int)
        wei_mat = np.tile(wei_list, (nrow, 1))
        sino_360[:, :overlap_int] = sino_360[:, :overlap_int] * wei_mat
        pad_wid = ncol - overlap_int
        sino_pad = np.pad(sino_360, ((0, 0), (pad_wid, 0)), mode='edge')
    cor = (sino_pad.shape[1] - 1.0) / 2.0
    return sino_pad, cor


def generate_sinogram_helical_scan(index, tomo_data, num_proj, pixel_size,
                                   y_start, y_stop, pitch, scan_type="180",
                                   angles=None, flat=None, dark=None,
                                   mask=None, crop=(0, 0, 0, 0)):
    """
    Generate a 180-degree sinogram or a 360-degree sinogram from a helical
    scan dataset which is a hdf/nxs object (Ref. [1]).

    Parameters
    ----------
    index : int
        Index of the sinogram.
    tomo_data : hdf object.
        3D array.
    num_proj : int
        Number of projections per 180-degree.
    pixel_size : float
        Pixel size. The unit must be the same as y-position.
    y_start : float
        Y-position of the stage at the beginning of the scan.
    y_stop : float
        Y-position of the stage at the end of the scan.
    pitch : float
        The distance which the y-stage is translated in one full rotation.
    scan_type : {"180", "360"}
        One of two options: "180" for generating a 180-degree sinogram or
        "360" for generating a 360-degree sinogram.
    angles : array_like, optional
        1D array. List of angles (degree) corresponding to acquired projections.
    flat : array_like, optional
        Flat-field image used for flat-field correction.
    dark : array_like, optional
        Dark-field image used for flat-field correction.
    mask : array_like, optional
        Used for removing streak artifacts caused by blobs in the flat-field
        image.
    crop : tuple of int, optional
        Used for cropping images.

    Returns
    -------
    sinogram : array_like
        2D array. 180-degree sinogram or 360-degree sinogram.
    list_angle : array_like
        1D array. List of angles corresponding to the generated sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    max_index = calc.calculate_maximum_index(y_start, y_stop, pitch,
                                             pixel_size, scan_type)
    (y_s, y_e) = calc.calculate_reconstructable_height(y_start, y_stop,
                                                       pitch, scan_type)
    if index < 0 or index > max_index:
        msg1 = "Requested index {0} is out of available index-range" \
               " [0, {1}]\n".format(index, max_index)
        msg2 = "corresponding to reconstructable heights" \
               " [{0}, {1}]".format(y_s, y_e)
        raise ValueError(msg1 + msg2)
    (depth0, height0, width0) = tomo_data.shape
    (crop_top, crop_bottom, crop_left, crop_right) = crop
    top = crop_top
    bottom = height0 - crop_bottom
    left = crop_left
    right = width0 - crop_right
    width = right - left
    height = bottom - top
    if flat is None:
        flat = np.ones((height0, width0), dtype=np.float32)
    if dark is None:
        dark = np.zeros((height0, width0), dtype=np.float32)
    if angles is None:
        step_angle = 180.0 / (num_proj - 1)
        angles = np.arange(0, depth0) * step_angle
    flat_dark = flat - dark
    FoV = pixel_size * height
    y_step = pitch / (2.0 * (num_proj - 1))
    if scan_type == "180":
        num_proj_used = num_proj
    else:
        num_proj_used = 2 * (num_proj - 1) + 1
    y_pos = (index - 1) * pixel_size + y_s
    i0 = int(np.ceil((y_e - y_pos) / y_step))
    if (i0 < 0) or (i0 >= depth0):
        raise ValueError(
            "Sinogram index {0} requests a projection index {1}"
            " which is out of the data range [0, {2}]".format(
                index, i0, depth0 - 1))
    sinogram = np.zeros((num_proj_used, width), dtype=np.float32)
    for i in range(i0, i0 + num_proj_used):
        j0 = (y_e + FoV - i * y_step - y_pos) / pixel_size - 1
        if (j0 < 0) or (j0 >= height):
            raise ValueError(
                "Requested row index {0} of projection {1} is out of the"
                " range [0, {2}]".format(j0, i0, height - 1))
        j0 = np.clip(j0, 0, height - 1)
        jd = int(np.floor(j0))
        ju = int(np.ceil(j0))
        list_down = (tomo_data[i, jd + crop_top, left: right]
                     - dark[jd + crop_top, left: right]) / flat_dark[
                                                           jd + crop_top,
                                                           left: right]
        if mask is not None:
            list_down = remo.remove_blob_1d(list_down,
                                            mask[jd + crop_top, left: right])
        if ju != jd:
            list_up = (tomo_data[i, ju + crop_top, left: right]
                       - dark[ju + crop_top, left: right]) \
                      / flat_dark[ju + crop_top, left: right]
            if mask is not None:
                list_up = remo.remove_blob_1d(list_up,
                                              mask[ju + crop_top, left: right])
            sinogram[i - i0] = list_down * (ju - j0) / (ju - jd) + list_up * (
                        j0 - jd) / (ju - jd)
        else:
            sinogram[i - i0] = list_down
    list_angle = angles[i0:i0 + num_proj_used]
    return sinogram, list_angle


def generate_full_sinogram_helical_scan(index, tomo_data, num_proj, pixel_size,
                                        y_start, y_stop, pitch, scan_type="180",
                                        angles=None, flat=None, dark=None,
                                        mask=None, crop=(0, 0, 0, 0)):
    """
    Generate a full sinogram from a helical scan dataset which is a hdf/nxs
    object (Ref. [1]). Full sinogram is all 1D projection of the same slice of
    a sample staying inside the field of view.

    Parameters
    ----------
    index : int
        Index of the sinogram.
    tomo_data : hdf object.
        3D array.
    num_proj : int
        Number of projections per 180-degree.
    pixel_size : float
        Pixel size. The unit must be the same as y-position.
    y_start : float
        Y-position of the stage at the beginning of the scan.
    y_stop : float
        Y-position of the stage at the end of the scan.
    pitch : float
        The distance which the y-stage is translated in one full rotation.
    scan_type : {"180", "360"}
        Data acquired is the 180-degree type or 360-degree type [1].
    angles : array_like, optional
        1D array. List of angles (degree) corresponding to acquired projections.
    flat : array_like, optional
        Flat-field image used for flat-field correction.
    dark : array_like, optional
        Dark-field image used for flat-field correction.
    mask : array_like, optional
        Used for removing streak artifacts caused by blobs in the flat-field
        image.
    crop : tuple of int, optional
        Used for cropping images.

    Returns
    -------
    sinogram : array_like
        2D array. Full sinogram.
    list_angle : array_like
        1D array. List of angles corresponding to the generated sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    (depth0, height0, width0) = tomo_data.shape
    (crop_top, crop_bottom, crop_left, crop_right) = crop
    top = crop_top
    bottom = height0 - crop_bottom
    left = crop_left
    right = width0 - crop_right
    width = right - left
    height = bottom - top
    if flat is None:
        flat = np.ones((height0, width0), dtype=np.float32)
    if dark is None:
        dark = np.zeros((height0, width0), dtype=np.float32)
    if angles is None:
        step_angle = 180.0 / (num_proj - 1)
        angles = np.arange(0, depth0) * step_angle
    flat_dark = flat - dark
    FoV = pixel_size * height
    y_step = pitch / (2.0 * (num_proj - 1))
    if scan_type == "180":
        y_e = y_stop - pitch / 2.0
        y_s = y_start + pitch / 2.0
    else:
        y_e = y_stop - pitch
        y_s = y_start + pitch
    num_proj_used = int(np.floor(FoV / y_step)) - 1
    y_pos = (index - 1) * pixel_size + y_s
    i0 = int(np.ceil((y_e - y_pos) / y_step))
    if (i0 < 0) or (i0 >= depth0):
        raise ValueError(
            "Sinogram index {0} requests a projection index {1} which "
            "is out of the projection range [0, {2}]".format(
                index, i0, depth0 - 1))
    if (i0 + num_proj_used) >= depth0:
        raise ValueError(
            "Sinogram index {0} requests projection-indices in the range of "
            "[{1}, {2}] which is out of the data range [0, {3}]".format(
                index, i0, i0 + num_proj_used, depth0 - 1))
    sinogram = np.zeros((num_proj_used, width), dtype=np.float32)
    for i in range(i0, i0 + num_proj_used):
        j0 = (y_e + FoV - i * y_step - y_pos) / pixel_size - 1
        if (j0 < 0) or (j0 >= height):
            raise ValueError(
                "Requested row index {0} of projection {1} is out of"
                " the range [0, {2}]".format(j0, i0, height))
        j0 = np.clip(j0, 0, height - 1)
        jd = int(np.floor(j0))
        ju = int(np.ceil(j0))
        list_down = (tomo_data[i, jd + crop_top, left: right]
                     - dark[jd + crop_top, left: right]) / flat_dark[
                                                           jd + crop_top,
                                                           left: right]
        if mask is not None:
            list_down = remo.remove_blob_1d(list_down,
                                            mask[jd + crop_top, left: right])
        if ju != jd:
            list_up = (tomo_data[i, ju + crop_top, left: right]
                       - dark[ju + crop_top, left: right]) / flat_dark[
                                                             ju + crop_top,
                                                             left: right]
            if mask is not None:
                list_up = remo.remove_blob_1d(list_up,
                                              mask[ju + crop_top, left: right])
            sinogram[i - i0] = list_down * (ju - j0) / (ju - jd) + list_up * (
                        j0 - jd) / (ju - jd)
        else:
            sinogram[i - i0] = list_down
    list_angle = angles[i0:i0 + num_proj_used]
    return sinogram, list_angle



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python implementations of preprocessing techniques.
# Contributors:
# ============================================================================

"""
Module of calculation methods in the preprocessing stage:
- Calculating the center-of-rotation (COR) in a 180-degree scan using a
  sinogram.
- Determining the overlap-side and overlap-area between images.
- Calculating the COR in a half-acquisition scan (360-degree scan with offset
  COR).
- Using the similar technique as above to calculate the COR in a 180-degree scan
  from two projections.
- Determining the relative translations between images using phase-correlation
  technique.
- Calculating the COR in a 180-degree scan using phase-correlation technique.
"""

import numpy as np
from scipy import stats
import scipy.ndimage as ndi
import multiprocessing as mp
from joblib import Parallel, delayed
import numpy.fft as fft


def make_inverse_double_wedge_mask(height, width, radius):
    """
    Generate a double-wedge binary mask using Eq. (3) in Ref. [1].
    Values outside the double-wedge region correspond to 1.0.

    Parameters
    ----------
    height : int
        Image height.
    width : int
        Image width.
    radius : int
        Radius of an object, in pixel unit.

    Returns
    -------
    array_like
        2D binary mask.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.22.019078
    """
    du = 1.0 / width
    dv = (height - 1.0) / (height * 2.0 * np.pi)
    ndrop = min(20, np.int16(0.05 * height))
    ycenter = np.int16(np.ceil((height - 1) / 2.0))
    xcenter = np.int16(np.ceil((width - 1) / 2.0))
    mask = np.zeros((height, width), dtype=np.float32)
    for i in range(height):
        num = np.int16(np.ceil(((i - ycenter) * dv / radius) / du))
        (p1, p2) = np.int16(np.clip(
            np.sort((-num + xcenter, num + xcenter)), 0, width - 1))
        mask[i, p1:p2 + 1] = 1.0
    mask[ycenter - ndrop:ycenter + ndrop + 1, :] = 0.0
    mask[:, xcenter - 1:xcenter + 2] = 0.0
    return mask


def calculate_center_metric(center, sino_180, sino_flip, sino_comp, mask):
    """
    Calculate a metric of an estimated center-of-rotation.

    Parameters
    ----------
    center : float
        Estimated center.
    sino_180 : array_like
        2D array. 180-degree sinogram.
    sino_flip : array_like
        2D array. Flip the 180-degree sinogram in the left/right direction.
    sino_comp : array_like
        2D array. Used to fill the gap left by image shifting.
    mask : array_like
        2D array. Used to select coefficients in the double-wedge region.

    Returns
    -------
    float
        Metric.
    """
    ncol = sino_180.shape[1]
    center_flip = (ncol - 1.0) / 2.0
    shift_col = 2.0 * (center - center_flip)
    if np.abs(shift_col - np.floor(shift_col)) == 0.0:
        shift_col = int(shift_col)
        sino_shift = np.roll(sino_flip, shift_col, axis=1)
        if shift_col >= 0:
            sino_shift[:, :shift_col] = sino_comp[:, :shift_col]
        else:
            sino_shift[:, shift_col:] = sino_comp[:, shift_col:]
        mat = np.vstack((sino_180, sino_shift))
    else:
        sino_shift = ndi.shift(sino_flip, (0, shift_col), order=3,
                               prefilter=True)
        if shift_col >= 0:
            shift_int = int(np.ceil(shift_col))
            sino_shift[:, :shift_int] = sino_comp[:, :shift_int]
        else:
            shift_int = int(np.floor(shift_col))
            sino_shift[:, shift_int:] = sino_comp[:, shift_int:]
        mat = np.vstack((sino_180, sino_shift))
    metric = np.mean(
        np.abs(np.fft.fftshift(fft.fft2(mat))) * mask)
    return metric


def coarse_search_cor(sino_180, start, stop, ratio=0.5, denoise=True,
                      ncore=None):
    """
    Find the center-of-rotation (COR) using integer shifting.

    Parameters
    ----------
    sino_180 : array_like
        2D array. 180-degree sinogram.
    start : int
        Starting point for searching COR.
    stop : int
        Ending point for searching COR.
    ratio : float
        Ratio between a sample and the width of the sinogram.
    denoise : bool, optional
        Apply a smoothing filter.
    ncore: int or None
        Number of cores used for computing. Automatically selected if None.

    Returns
    -------
    float
        Center of rotation.
    """
    if denoise is True:
        sino_180 = ndi.gaussian_filter(sino_180, (3, 1), mode='reflect')
    if ncore is None:
        ncore = np.clip(mp.cpu_count() - 1, 1, None)
    (nrow, ncol) = sino_180.shape
    start_cor = np.int16(np.clip(start, 0, ncol - 1))
    stop_cor = np.int16(np.clip(stop, 0, ncol - 1))
    sino_flip = np.fliplr(sino_180)
    sino_comp = np.flipud(sino_180)
    list_cor = np.arange(start_cor, stop_cor + 1)
    list_metric = np.zeros(len(list_cor), dtype=np.float32)
    num_metric = len(list_metric)
    mask = make_inverse_double_wedge_mask(2 * nrow, ncol, 0.5 * ratio * ncol)
    if ncore == 1:
        for i, cor in enumerate(list_cor):
            list_metric[i] = calculate_center_metric(
                list_cor[i], sino_180, sino_flip, sino_comp, mask)
    else:
        list_metric = Parallel(n_jobs=ncore, backend="threading")(
            delayed(calculate_center_metric)(list_cor[i], sino_180, sino_flip,
                                             sino_comp, mask) for i in
            range(num_metric))
    return list_cor[np.argmin(list_metric)]


def fine_search_cor(sino_180, start, radius, step, ratio=0.5, denoise=True,
                    ncore=None):
    """
    Find the center-of-rotation (COR) using sub-pixel shifting.

    Parameters
    ----------
    sino_180 : array_like
        2D array. 180-degree sinogram.
    start : float
        Starting point for searching COR.
    radius : float
        Searching range: [start - radius; start + radius].
    step : float
        Searching step.
    ratio : float
        Ratio between a sample and the width of the sinogram.
    denoise : bool, optional
        Apply a smoothing filter.
    ncore: int or None
        Number of cores used for computing. Automatically selected if None.

    Returns
    -------
    float
        Center of rotation.
    """
    if denoise is True:
        sino_180 = ndi.gaussian_filter(sino_180, (2, 2), mode='reflect')
    if ncore is None:
        ncore = np.clip(mp.cpu_count() - 1, 1, None)
    (nrow, ncol) = sino_180.shape
    sino_flip = np.fliplr(sino_180)
    sino_comp = np.flipud(sino_180)
    list_cor = np.clip(
        start + np.arange(-radius, radius + step, step), 0.0, ncol - 1.0)
    list_metric = np.zeros(len(list_cor), dtype=np.float32)
    num_metric = len(list_metric)
    mask = make_inverse_double_wedge_mask(2 * nrow, ncol, 0.5 * ratio * ncol)
    if ncore == 1:
        for i, cor in enumerate(list_cor):
            list_metric[i] = calculate_center_metric(
                list_cor[i], sino_180, sino_flip, sino_comp, mask)
    else:
        list_metric = Parallel(n_jobs=ncore, backend="threading")(
            delayed(calculate_center_metric)(list_cor[i], sino_180, sino_flip,
                                             sino_comp, mask) for i in
            range(num_metric))
    return list_cor[np.argmin(list_metric)]


def downsample_cor(image, dsp_fact0, dsp_fact1):
    """
    Downsample an image by averaging.

    Parameters
    ----------
    image : array_like
        2D array.
    dsp_fact0 : int
        Downsampling factor along axis 0.
    dsp_fact1 : int
        Downsampling factor along axis 1.

    Returns
    -------
    array_like
        2D array. Downsampled image.
    """
    (height, width) = image.shape
    dsp_fact0 = np.clip(np.int16(dsp_fact0), 1, height // 2)
    dsp_fact1 = np.clip(np.int16(dsp_fact1), 1, width // 2)
    height_dsp = height // dsp_fact0
    width_dsp = width // dsp_fact1
    image_dsp = image[0:dsp_fact0 * height_dsp, 0:dsp_fact1 * width_dsp]
    image_dsp = image_dsp.reshape(
        height_dsp, dsp_fact0, width_dsp, dsp_fact1).mean(-1).mean(1)
    return image_dsp


def find_center_vo(sino_180, start=None, stop=None, step=0.25, radius=4,
                   ratio=0.5, dsp=True, ncore=None):
    """
    Find the center-of-rotation using the method described in Ref. [1].

    Parameters
    ----------
    sino_180 : array_like
        2D array. 180-degree sinogram.
    start : float
        Starting point for searching CoR.
    stop : float
        Ending point for searching CoR.
    step : float
        Sub-pixel accuracy of estimated CoR.
    radius : float
        Searching range with the sub-pixel step.
    ratio : float
        Ratio between the sample and the width of the sinogram.
    dsp : bool
        Enable/disable downsampling.
    ncore: int or None
        Number of cores used for computing. Automatically selected if None.

    Returns
    -------
    float
        Center-of-rotation.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.22.019078
    """
    (nrow, ncol) = sino_180.shape
    if start is None:
        start = ncol // 2 - ncol // 16
    if stop is None:
        stop = ncol // 2 + ncol // 16
    dsp_row = 1
    dsp_col = 1
    if dsp is True:
        if ncol > 1000:
            dsp_col = int(np.floor(ncol / 640.0))
        if nrow > 1000:
            dsp_row = int(np.floor(nrow / 900.0))
        sino_dns = ndi.gaussian_filter(sino_180, (3, 1), mode='reflect')
        sino_dsp = downsample_cor(sino_dns, dsp_row, dsp_col)
        radius = max(radius, dsp_col)
        off_set = 0.5 * dsp_col
        start = int(np.floor(1.0 * start / dsp_col))
        stop = int(np.ceil(1.0 * stop / dsp_col))
        raw_cor = coarse_search_cor(sino_dsp, start, stop, ratio,
                                    denoise=False, ncore=ncore)
        fine_cor = fine_search_cor(sino_180, raw_cor * dsp_col + off_set,
                                   radius, step, ratio, denoise=True,
                                   ncore=ncore)
    else:
        raw_cor = coarse_search_cor(sino_180, start, stop, ratio, denoise=True,
                                    ncore=ncore)
        fine_cor = fine_search_cor(sino_180, raw_cor, radius, step, ratio,
                                   denoise=True, ncore=ncore)
    return fine_cor


def calculate_curvature(list_metric):
    """
    Calculate the curvature of a fitted curve going through the minimum
    value of a metric list.

    Parameters
    ----------
    list_metric : array_like
        1D array. List of metrics.

    Returns
    -------
    curvature : float
        Quadratic coefficient of the parabola fitting.
    min_pos : float
        Position of the minimum value with sub-pixel accuracy.
    """
    radi = 2
    num_metric = len(list_metric)
    min_pos = np.clip(
        np.argmin(list_metric), radi, num_metric - radi - 1)
    list1 = list_metric[min_pos - radi:min_pos + radi + 1]
    (afact1, _, _) = np.polyfit(np.arange(0, 2 * radi + 1), list1, 2)
    list2 = list_metric[min_pos - 1:min_pos + 2]
    (afact2, bfact2, _) = np.polyfit(
        np.arange(min_pos - 1, min_pos + 2), list2, 2)
    curvature = np.abs(afact1)
    if afact2 != 0.0:
        num = - bfact2 / (2 * afact2)
        if (num >= min_pos - 1) and (num <= min_pos + 1):
            min_pos = num
    return curvature, np.float32(min_pos)


def correlation_metric(mat1, mat2):
    """
    Calculate the correlation metric. Smaller metric corresponds to better
    correlation.

    Parameters
    ---------
    mat1 : array_like
    mat2 : array_like

    Returns
    -------
    float
        Correlation metric.
    """
    metric = np.abs(
        1.0 - stats.pearsonr(mat1.flatten('F'), mat2.flatten('F'))[0])
    return metric


def search_overlap(mat1, mat2, win_width, side, denoise=True, norm=False,
                   use_overlap=False):
    """
    Calculate the correlation metrics between a rectangular region, defined
    by the window width, on the utmost left/right side of image 2 and the
    same size region in image 1 where the region is slided across image 1.

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image or sinogram image.
    mat2 : array_like
        2D array. Projection image or sinogram image.
    win_width : int
        Width of the searching window.
    side : {0, 1}
        Only two options: 0 or 1. It is used to indicate the overlap side
        respects to image 1. "0" corresponds to the left side. "1" corresponds
        to the right side.
    denoise : bool, optional
        Apply the Gaussian filter if True.
    norm : bool, optional
        Apply the normalization if True.
    use_overlap : bool, optional
        Use the combination of images in the overlap area for calculating
        correlation coefficients if True.

    Returns
    -------
    list_metric : array_like
        1D array. List of the correlation metrics.
    offset : int
        Initial position of the searching window where the position
        corresponds to the center of the window.
    """
    if denoise is True:
        mat1 = ndi.gaussian_filter(mat1, (2, 2), mode='reflect')
        mat2 = ndi.gaussian_filter(mat2, (2, 2), mode='reflect')
    (nrow1, ncol1) = mat1.shape
    (nrow2, ncol2) = mat2.shape
    if nrow1 != nrow2:
        raise ValueError("Two images are not at the same height!!!")
    win_width = np.int16(np.clip(win_width, 6, min(ncol1, ncol2) // 2 - 1))
    offset = win_width // 2
    win_width = 2 * offset  # Make it even
    ramp_down = np.linspace(1.0, 0.0, win_width)
    ramp_up = 1.0 - ramp_down
    wei_down = np.tile(ramp_down, (nrow1, 1))
    wei_up = np.tile(ramp_up, (nrow1, 1))
    if side == 1:
        mat2_roi = mat2[:, 0:win_width]
        mat2_roi_wei = mat2_roi * wei_up
    else:
        mat2_roi = mat2[:, ncol2 - win_width:]
        mat2_roi_wei = mat2_roi * wei_down
    list_mean2 = np.mean(np.abs(mat2_roi), axis=1)
    list_pos = np.arange(offset, ncol1 - offset)
    num_metric = len(list_pos)
    list_metric = np.ones(num_metric, dtype=np.float32)
    for i, pos in enumerate(list_pos):
        mat1_roi = mat1[:, pos - offset:pos + offset]
        if use_overlap is True:
            if side == 1:
                mat1_roi_wei = mat1_roi * wei_down
            else:
                mat1_roi_wei = mat1_roi * wei_up
        if norm is True:
            list_mean1 = np.mean(np.abs(mat1_roi), axis=1)
            list_fact = list_mean2 / list_mean1
            mat_fact = np.transpose(np.tile(list_fact, (win_width, 1)))
            mat1_roi = mat1_roi * mat_fact
            if use_overlap is True:
                mat1_roi_wei = mat1_roi_wei * mat_fact
        if use_overlap is True:
            mat_comb = mat1_roi_wei + mat2_roi_wei
            list_metric[i] = (correlation_metric(mat1_roi, mat2_roi)
                              + correlation_metric(mat1_roi, mat_comb)
                              + correlation_metric(mat2_roi, mat_comb)) / 3.0
        else:
            list_metric[i] = correlation_metric(mat1_roi, mat2_roi)
    min_metric = np.min(list_metric)
    if min_metric != 0.0:
        list_metric = list_metric / min_metric
    return list_metric, offset


def find_overlap(mat1, mat2, win_width, side=None, denoise=True, norm=False,
                 use_overlap=False):
    """
    Find the overlap area and overlap side between two images (Ref. [1]) where
    the overlap side referring to the first image.

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image or sinogram image.
    mat2 :  array_like
        2D array. Projection image or sinogram image.
    win_width : int
        Width of the searching window.
    side : {None, 0, 1}, optional
        Only there options: None, 0, or 1. "None" corresponding to fully
        automated determination. "0" corresponding to the left side. "1"
        corresponding to the right side.
    denoise : bool, optional
        Apply the Gaussian filter if True.
    norm : bool, optional
        Apply the normalization if True.
    use_overlap : bool, optional
        Use the combination of images in the overlap area for calculating
        correlation coefficients if True.

    Returns
    -------
    overlap : float
        Width of the overlap area between two images.
    side : int
        Overlap side between two images.
    overlap_position : float
        Position of the window in the first image giving the best
        correlation metric.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    (_, ncol1) = mat1.shape
    (_, ncol2) = mat2.shape
    win_width = np.int16(np.clip(win_width, 6, min(ncol1, ncol2) // 2))
    if side == 1:
        (list_metric, offset) = search_overlap(mat1, mat2, win_width, side,
                                               denoise, norm, use_overlap)
        (_, overlap_position) = calculate_curvature(list_metric)
        overlap_position = overlap_position + offset
        overlap = ncol1 - overlap_position + win_width // 2
    elif side == 0:
        (list_metric, offset) = search_overlap(mat1, mat2, win_width, side,
                                               denoise, norm, use_overlap)
        (_, overlap_position) = calculate_curvature(list_metric)
        overlap_position = overlap_position + offset
        overlap = overlap_position + win_width // 2
    else:
        (list_metric1, offset1) = search_overlap(mat1, mat2, win_width, 1, norm,
                                                 denoise, use_overlap)
        (list_metric2, offset2) = search_overlap(mat1, mat2, win_width, 0, norm,
                                                 denoise, use_overlap)
        (curvature1, overlap_position1) = calculate_curvature(list_metric1)
        overlap_position1 = overlap_position1 + offset1
        (curvature2, overlap_position2) = calculate_curvature(list_metric2)
        overlap_position2 = overlap_position2 + offset2
        if curvature1 > curvature2:
            side = 1
            overlap_position = overlap_position1
            overlap = ncol1 - overlap_position + win_width // 2
        else:
            side = 0
            overlap_position = overlap_position2
            overlap = overlap_position + win_width // 2
    return overlap, side, overlap_position


def find_overlap_multiple(list_mat, win_width, side=None, denoise=True,
                          norm=False, use_overlap=False):
    """
    Find the overlap-areas and overlap-sides of a list of images where the
    overlap side referring to the previous image.

    Parameters
    ----------
    list_mat : list of array_like
        List of 2D array. Projection image or sinogram image.
    win_width : int
        Width of the searching window.
    side : {None, 0, 1}, optional
        Only there options: None, 0, or 1. "None" corresponding to fully
        automated determination. "0" corresponding to the left side. "1"
        corresponding to the right side.
    denoise : bool, optional
        Apply the Gaussian filter if True.
    norm : bool, optional
        Apply the normalization if True.
    use_overlap : bool, optional
        Use the combination of images in the overlap area for calculating
        correlation coefficients if True.

    Returns
    -------
    list_overlap : list of tuple of floats
        List of [overlap, side, overlap_position].
        overlap : Width of the overlap area between two images.
        side : Overlap side between two images.
        overlap_position : Position of the window in the first
        image giving the best correlation metric.
    """
    list_overlap = []
    num_mat = len(list_mat)
    if num_mat > 1:
        for i in range(num_mat-1):
            results = find_overlap(list_mat[i], list_mat[i + 1], win_width,
                                   side, denoise, norm, use_overlap)
            list_overlap.append(results)
    else:
        raise ValueError("Need at least 2 images to work!!!")
    return list_overlap


def find_center_360(sino_360, win_width, side=None, denoise=True, norm=False,
                    use_overlap=False):
    """
    Find the center-of-rotation (COR) in a 360-degree scan with offset COR use
    the method presented in Ref. [1].

    Parameters
    ----------
    sino_360 : array_like
        2D array. 360-degree sinogram.
    win_width : int
        Window width used for finding the overlap area.
    side : {None, 0, 1}, optional
        Overlap size. Only there options: None, 0, or 1. "None" corresponding
        to fully automated determination. "0" corresponding to the left side.
        "1" corresponding to the right side.
    denoise : bool, optional
        Apply the Gaussian filter if True.
    norm : bool, optional
        Apply the normalization if True.
    use_overlap : bool, optional
        Use the combination of images in the overlap area for calculating
        correlation coefficients if True.

    Returns
    -------
    cor : float
        Center-of-rotation.
    overlap : float
        Width of the overlap area between two halves of the sinogram.
    side : int
        Overlap side between two halves of the sinogram.
    overlap_position : float
        Position of the window in the first image giving the best
        correlation metric.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    (nrow, ncol) = sino_360.shape
    nrow_180 = nrow // 2 + 1
    sino_top = sino_360[0:nrow_180, :]
    sino_bot = np.fliplr(sino_360[-nrow_180:, :])
    (overlap, side, overlap_position) = find_overlap(
        sino_top, sino_bot, win_width, side, denoise, norm, use_overlap)
    if side == 0:
        cor = overlap / 2.0 - 1.0
    else:
        cor = ncol - overlap / 2.0 - 1.0
    return cor, overlap, side, overlap_position


def complex_gradient(mat):
    """
    Return complex gradient of a 2D array.
    """
    mat1a = np.roll(mat, -2, axis=1)
    mat2a = mat1a - mat
    mat2a[:, :2] = 0.0
    mat2a[:, -2:] = 0.0
    mat1b = np.roll(mat, -2, axis=0)
    mat2b = mat1b - mat
    mat2b[:2] = 0.0
    mat2b[-2:] = 0.0
    mat2 = mat2a + 1j * mat2b
    return mat2


def find_shift_based_phase_correlation(mat1, mat2, gradient=True):
    """
    Find relative translation in x and y direction between images with
    haft-pixel accuracy (Ref. [1]).

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image or sinogram image.
    mat2 : array_like
        2D array. Projection image or sinogram image.
    gradient : bool, optional
        Use the complex gradient of the input image for calculation.

    Returns
    -------
    ty : float
        Translation in y-direction.
    tx : float
        Translation in x-direction.

    References
    ----------
    .. [1] https://doi.org/10.1049/el:20030666
    """
    if gradient is True:
        mat1 = complex_gradient(mat1)
        mat2 = complex_gradient(mat2)
    (nrow, ncol) = mat1.shape
    mat_tmp1 = fft.fft2(mat1) * np.conjugate(fft.fft2(mat2))
    mat_tmp2 = np.abs(mat_tmp1)
    mat_tmp2[mat_tmp2 == 0.0] = 1.0
    mat3 = np.abs(fft.ifft2(mat_tmp1 / mat_tmp2))
    (ty, tx) = np.unravel_index(np.argmax(mat3, axis=None), mat3.shape)
    list_x = np.asarray([tx - 1, tx, tx + 1])
    list_vx = np.asarray(
        [mat3[ty, tx - 1], mat3[ty, tx], mat3[ty, (tx + 1) % ncol]])
    list_y = np.asarray([ty - 1, ty, ty + 1])
    list_vy = np.asarray(
        [mat3[ty - 1, tx], mat3[ty, tx], mat3[(ty + 1) % nrow, tx]])
    (afact_x, bfact_x, _) = np.polyfit(list_x, list_vx, 2)
    (afact_y, bfact_y, _) = np.polyfit(list_y, list_vy, 2)
    if afact_x != 0.0:
        num = - bfact_x / (2 * afact_x)
        if (num >= tx - 1) and (num <= tx + 1):
            tx = num
    if afact_y != 0.0:
        num = - bfact_y / (2 * afact_y)
        if (num >= ty - 1) and (num <= ty + 1):
            ty = num
    xcenter = np.ceil((ncol - 1) * 0.5)
    ycenter = np.ceil((nrow - 1) * 0.5)
    if ty > ycenter:
        ty = -(nrow - ty)
    if tx > xcenter:
        tx = -(ncol - tx)
    return ty, tx


def find_center_based_phase_correlation(mat1, mat2, flip=True, gradient=True):
    """
    Find the center-of-rotation (COR) using projection images at 0-degree
    and 180-degree.

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image at 0-degree.
    mat2 : array_like
        2D array. Projection image at 180-degree.
    flip : bool, optional
        Flip the 180-degree projection in the left-right direction if True.
    gradient : bool, optional
        Use the complex gradient of the input image for calculation.

    Returns
    -------
    cor : float
        Center-of-rotation.
    """
    ncol = mat1.shape[-1]
    if flip is True:
        mat2 = np.fliplr(mat2)
    tx = find_shift_based_phase_correlation(mat1, mat2, gradient=gradient)[-1]
    cor = (ncol - 1.0 + tx) * 0.5
    return cor


def find_center_projection(mat1, mat2, flip=True, chunk_height=None,
                           start_row=None, denoise=True, norm=False,
                           use_overlap=False):
    """
    Find the center-of-rotation (COR) using projection images at 0-degree
    and 180-degree based on a method in Ref. [1].

    Parameters
    ----------
    mat1 : array_like
        2D array. Projection image at 0-degree.
    mat2 : array_like
        2D array. Projection image at 180-degree.
    flip : bool, optional
        Flip the 180-degree projection in the left-right direction if True.
    chunk_height : int or float, optional
        Height of the sub-area of projection images. If a float is given, it
        must be in the range of [0.0, 1.0].
    start_row : int, optional
        Starting row used to extract the sub-area.
    denoise : bool, optional
        Apply the Gaussian filter if True.
    norm : bool, optional
        Apply the normalization if True.
    use_overlap : bool, optional
        Use the combination of images in the overlap area for calculating
        correlation coefficients if True.

    Returns
    -------
    cor : float
        Center-of-rotation.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    (nrow, ncol) = mat1.shape
    if flip is True:
        mat2 = np.fliplr(mat2)
    win_width = ncol // 2
    if chunk_height is None:
        chunk_height = int(0.1 * nrow)
    if isinstance(chunk_height, float):
        if 0.0 < chunk_height <= 1.0:
            chunk_height = int(chunk_height * nrow)
        else:
            chunk_height = int(0.1 * nrow)
    chunk_height = np.clip(chunk_height, 1, nrow - 1)
    if start_row is None:
        start = nrow // 2 - chunk_height // 2
    elif start_row < 0:
        start = nrow + start_row - chunk_height // 2
    else:
        start = start_row - chunk_height // 2
    stop = start + chunk_height
    start = np.clip(start, 0, nrow - chunk_height - 1)
    stop = np.clip(stop, chunk_height, nrow - 1)
    mat1_roi = mat1[start: stop]
    mat2_roi = mat2[start: stop]
    (overlap, side, _) = find_overlap(mat1_roi, mat2_roi, win_width, side=None,
                                      denoise=denoise, norm=norm,
                                      use_overlap=use_overlap)
    if side == 0:
        cor = overlap / 2.0 - 1.0
    else:
        cor = ncol - overlap / 2.0 - 1.0
    return cor


def calculate_reconstructable_height(y_start, y_stop, pitch, scan_type):
    """
    Calculate reconstructable height in a helical scan.

    Parameters
    ----------
    y_start : float
        Y-position of the stage at the beginning of the scan.
    y_stop : float
        Y-position of the stage at the end of the scan.
    pitch : float
        The distance which the y-stage is translated in one full rotation.
    scan_type : {"180", "360"}
        One of two options: "180" for generating a 180-degree sinogram or
        "360" for generating a 360-degree sinogram.

    Returns
    -------
    y_s : float
        Starting point of the reconstructable height.
    y_e : float
        End point of the reconstructable height.
    """
    if not(scan_type == "180" or scan_type == "360"):
        raise ValueError("!!! Please one of two options: '180' or '360'!!!")
    if scan_type == "360":
        y_s = y_start + pitch
        y_e = y_stop - pitch
    else:
        y_s = y_start + pitch / 2.0
        y_e = y_stop - pitch / 2.0
    return y_s, y_e


def calculate_maximum_index(y_start, y_stop, pitch, pixel_size, scan_type):
    """
    Calculate the maximum index of a reconstructable slice in a helical scan.

    Parameters
    ----------
    y_start : float
        Y-position of the stage at the beginning of the scan.
    y_stop : float
        Y-position of the stage at the end of the scan.
    pitch : float
        The distance which the y-stage is translated in one full rotation.
    pixel_size : float
        Pixel size. The unit must be the same as y-position.
    scan_type : {"180", "360"}
        One of two options: "180" for generating a 180-degree sinogram or
        "360" for generating a 360-degree sinogram.

    Returns
    -------
    int
        Maximum index of reconstructable slices.
    """
    if not(scan_type == "180" or scan_type == "360"):
        raise ValueError("!!! Please one of two options: '180' or '360'!!!")
    y_s, y_e = calculate_reconstructable_height(y_start, y_stop, pitch,
                                                scan_type)
    max_index = int(((y_e - y_s) / pixel_size)) + 1
    return max_index



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Utility methods
# Contributors:
# ============================================================================

"""
Module of utility methods:
1- Methods for parallel computing, geometric transformation, masking.
2- Methods for customizing stripe/ring removal methods
    2.1  - sort_forward
    2.2  - sort_backward
    2.3  - separate_frequency_component
    2.4  - generate_fitted_image
    2.5  - detect_stripe
    2.6  - calculate_regularization_coefficient
    2.7  - make_2d_butterworth_window
    2.8  - make_2d_damping_window
    2.9  - apply_wavelet_decomposition
    2.10 - apply_wavelet_recontruction
    2.11 - apply_filter_to_wavelet_component
    2.12 - interpolate_inside_stripe
    2.13 - transform_slice_forward
    2.14 - transform_slice_backward
3- Customized smoothing filters:
    3.1 - apply_gaussian_filter (in the Fourier space)
    3.2 - apply_regularization_filter
4- Methods for grid scans:
    4.1  - detec_sample
    4.2  - fix_non_sample_areas
    4.3  - locate_slice
    4.4  - locate_slice_chunk
 """

import sys
import multiprocessing as mp
import pywt
import numpy as np
import scipy.ndimage as ndi
from scipy import interpolate
import scipy.signal.windows as win
from scipy.signal import savgol_filter
from joblib import Parallel, delayed
import numpy.fft as fft
import algotom.prep.removal as remo
import algotom.prep.filtering as filt
import algotom.rec.reconstruction as reco


def apply_method_to_multiple_sinograms(data, method, para, ncore=None):
    """
    Apply a processing method (in "filtering", "removal", and "reconstruction"
    module) to multiple sinograms or multiple slices in parallel.

    Parameters
    ----------
    data : array_like or hdf object
        3D array data where sinograms/slices are extracted along axis 1,
        e.g [:, i, :].
    method : str
        Name of a method. e.g. "remove_stripe_based_sorting".
    para : list
        Parameters of the method. e.g. [21, 1]
    ncore: int or None
        Number of cores used for computing. Automatically selected if None.

    Returns
    -------
    array_like
        Same axis-definition as the input.
    """
    if ncore is None:
        ncore = np.clip(mp.cpu_count() - 1, 1, None)
    else:
        ncore = np.clip(ncore, 1, None)
    if not isinstance(para, list):
        para = tuple(list([para]))
    else:
        para = tuple(para)
    (depth, height, width) = data.shape
    if method in dir(remo):
        method_used = getattr(remo, method)
    elif method in dir(filt):
        method_used = getattr(filt, method)
    elif method in dir(reco):
        method_used = getattr(reco, method)
    else:
        raise ValueError("Can't find the method: '{}' in the namespace"
                         "".format(method))
    data_out = Parallel(n_jobs=ncore, backend="threading")(
        delayed(method_used)(data[:, i, :], *para) for i in range(height))
    data_out = np.moveaxis(np.asarray(data_out), 0, 1)
    return data_out


def mapping(mat, xmat, ymat, order=1, mode="reflect"):
    """
    Apply a geometric transformation to a 2D array

    Parameters
    ----------
    mat : array_like
        2D array.
    xmat : array_like
        2D array of the x-coordinates.
    ymat : array_like
        2D array of the y-coordinates.
    order : int, optional
        The order of the spline interpolation, default is 1.
        The order has to be in the range 0-5.
    mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap'}, optional
        The mode parameter determines how the input array is extended beyond
        its boundaries. Default is 'reflect'.

    Returns
    -------
    array_like
        2D array.
    """
    coords = np.vstack((np.ndarray.flatten(ymat), np.ndarray.flatten(xmat)))
    mat = ndi.map_coordinates(mat, coords, order=order, mode=mode)
    return mat.reshape(xmat.shape)


def make_circle_mask(width, ratio):
    """
    Create a circle mask.

    Parameters
    -----------
    width : int
        Width of a square array.
    ratio : float
        Ratio between the diameter of the mask and the width of the array.

    Returns
    ------
    array_like
         Square array.
    """
    mask = np.zeros((width, width), dtype=np.float32)
    center = width // 2
    radius = ratio * center
    y, x = np.ogrid[-center:width - center, -center:width - center]
    mask_check = x * x + y * y <= radius * radius
    mask[mask_check] = 1.0
    return mask


def sort_forward(mat, axis=0):
    """
    Sort grayscales of an image along an axis.
    e.g. axis=0 is to sort along each column.

    Parameters
    ----------
    mat : array_like
        2D array.
    axis : int
        Axis along which to sort.

    Returns
    --------
    mat_sort : array_like
        2D array. Sorted image.
    mat_index : array_like
        2D array. Index array used for sorting backward.
    """
    if axis == 0:
        mat = np.transpose(mat)
    (nrow, ncol) = mat.shape
    list_index = np.arange(0.0, ncol, 1.0)
    mat_index = np.tile(list_index, (nrow, 1))
    mat_comb = np.asarray(np.dstack((mat_index, mat)))
    mat_comb_sort = np.asarray(
        [row[row[:, 1].argsort()] for row in mat_comb])
    mat_sort = mat_comb_sort[:, :, 1]
    mat_index = mat_comb_sort[:, :, 0]
    if axis == 0:
        mat_sort = np.transpose(mat_sort)
        mat_index = np.transpose(mat_index)
    return mat_sort, mat_index


def sort_backward(mat, mat_index, axis=0):
    """
    Sort grayscales of an image using an index array provided.
    e.g axis=0 is to sort each column.

    Parameters
    ----------
    mat : array_like
        2D array.
    mat_index : array_like
        2D array. Index array used for sorting.
    axis : int
        Axis along which to sort.

    Returns
    --------
    mat_sort : array_like
        2D array. Sorted image.
    """
    if axis == 0:
        mat = np.transpose(mat)
        mat_index = np.transpose(mat_index)
    mat_comb = np.asarray(np.dstack((mat_index, mat)))
    mat_comb_sort = np.asarray(
        [row[row[:, 0].argsort()] for row in mat_comb])
    mat_sort = mat_comb_sort[:, :, 1]
    if axis == 0:
        mat_sort = np.transpose(mat_sort)
    return mat_sort


def separate_frequency_component(mat, axis=0,
                                 window={"name": "gaussian", "sigma": 5}):
    """
    Separate low and high frequency components of an image along an axis.
    e.g axis=0 is to apply the separation to each column.

    Parameters
    ----------
    mat : array_like
        2D array.
    axis : int
        Axis along which to apply the filter.
    window : array_like or dict
        1D array or a dictionary which given the name of a window in
        the scipy.signal.window list and its parameters (without window-length).

    Returns
    -------
    mat_low : array_like
        2D array. Low-frequency image.
    mat_high : array_like
        2D array. High-frequency image.
    """
    if axis == 0:
        mat = np.transpose(mat)
    (nrow, ncol) = mat.shape
    pad = min(150, int(0.1 * ncol))
    if not isinstance(window, dict):
        if len(window) != ncol:
            raise ValueError("Window-length is not the same as the"
                             " axis-length!")
        else:
            window = np.pad(window, (pad, pad), mode='constant',
                            constant_values=1.0)
    else:
        win_name = tuple(window.values())[0]
        para = tuple(window.values())[1:]
        window = getattr(win, win_name)(ncol + 2 * pad, *para)

    list_sign = np.power(-1.0, np.arange(ncol + 2 * pad))
    mat_pad = np.pad(mat, ((0, 0), (pad, pad)), mode='reflect')
    mat_smooth = np.copy(mat)
    for i, line in enumerate(mat_pad):
        mat_smooth[i] = np.real(
            fft.ifft(fft.fft(line * list_sign) * window) *
            list_sign)[pad:ncol + pad]
    mat_sharp = mat - mat_smooth
    if axis == 0:
        mat_smooth = np.transpose(mat_smooth)
        mat_sharp = np.transpose(mat_sharp)
    return mat_smooth, mat_sharp


def generate_fitted_image(mat, order, axis=0, num_chunk=1):
    """
    Apply a polynomial fitting along an axis of an image.
    e.g. axis=0 is to apply the fitting to each column.

    Parameters
    ----------
    mat : array_like
        2D array.
    order : int
        Order of the polynomial used to fit.
    axis : int
        Axis along which to apply the filter.
    num_chunk : int
        Number of chunks of rows or columns to apply the fitting.

    Returns
    -------
    mat_fit : array_like
    """
    (nrow, ncol) = mat.shape
    if axis == 0:
        size = nrow // num_chunk
        length = size if (size % 2 == 1) else size - 1
        mat_fit = savgol_filter(mat, length, order, axis=axis, mode='mirror')
    else:
        size = ncol // num_chunk
        length = size if (size % 2 == 1) else size - 1
        mat_fit = savgol_filter(mat, length, order, axis=axis, mode='mirror')
    return mat_fit


def detect_stripe(list_data, snr):
    """
    Locate stripe positions using Algorithm 4 in Ref. [1]

    Parameters
    ----------
    list_data : array_like
        1D array. Normalized data.
    snr : float
        Ratio (>1.0) used to detect stripe locations. Greater is less sensitive.

    Returns
    -------
    array_like
        1D binary mask.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    npoint = len(list_data)
    list_sort = np.sort(list_data)
    xlist = np.arange(0, npoint, 1.0)
    ndrop = np.int16(0.25 * npoint)
    (slope, intercept) = np.polyfit(xlist[ndrop:-ndrop - 1],
                                    list_sort[ndrop:-ndrop - 1], 1)
    y_end = intercept + slope * xlist[-1]
    noise_level = np.abs(y_end - intercept)
    if noise_level == 0.0:
        raise ValueError("The method doesn't work on noise-free data. If you "
                         "apply the method on simulated data, please add"
                         " noise!")
    val1 = np.abs(list_sort[-1] - y_end) / noise_level
    val2 = np.abs(intercept - list_sort[0]) / noise_level
    list_mask = np.zeros(npoint, dtype=np.float32)
    if val1 >= snr:
        upper_thresh = y_end + noise_level * snr * 0.5
        list_mask[list_data > upper_thresh] = 1.0
    if val2 >= snr:
        lower_thresh = intercept - noise_level * snr * 0.5
        list_mask[list_data <= lower_thresh] = 1.0
    return list_mask


def calculate_regularization_coefficient(width, alpha):
    """
    Calculate coefficients used for the regularization-based method.
    Eq. (7) in Ref. [1].

    Parameters
    ----------
    width : int
        Width of a square array.
    alpha : float
        Regularization parameter.

    Returns
    -------
    float
         Square array.

    References
    ----------
    .. [1] https://doi.org/10.1016/j.aml.2010.08.022
    """
    tau = 2.0 * np.arcsinh(np.sqrt(alpha) * 0.5)
    ilist = np.arange(0, width)
    jlist = np.arange(0, width)
    matjj, matii = np.meshgrid(jlist, ilist)
    mat1 = np.abs(matii - matjj)
    mat2 = matii + matjj
    mat1a = np.cosh((width - 1 - mat1) * tau)
    mat2a = np.cosh((width - mat2) * tau)
    matcoe = - (np.tanh(
        0.5 * tau) / (alpha * np.sinh(width * tau))) * (mat1a + mat2a)
    return matcoe


def make_2d_butterworth_window(width, height, u, v, n):
    """
    Create a 2d window from the 1D Butterworth window.

    Parameters
    ----------
    height : int
        Height of the window.
    width : int
        Width of the window.
    u : int
        Cutoff frequency.
    n : int
        Filter order.
    v : int
        Number of rows (= 2*v) around the height middle are the
        1D Butterworth windows.

    Returns
    -------
    array_like
        2D array.
    """
    xcenter = np.ceil(width / 2.0) - 1.0
    ycenter = np.int16(np.ceil(height / 2.0) - 1)
    xlist = np.arange(width) - xcenter
    window = 1.0 / (1.0 + np.power(xlist / u, 2 * n))
    row1 = ycenter - np.int16(v)
    row2 = ycenter + np.int16(v) + 1
    window_2d = np.ones((height, width), dtype=np.float32)
    window_2d[row1:row2] = window
    return window_2d


def make_2d_damping_window(width, height, size, window_name="gaussian"):
    """
    Make 2D damping window from a list of 1D window for a Fourier-space filter,
    i.e. a high-pass filter.

    Parameters
    ----------
    height : int
        Height of the window.
    width : int
        Width of the window.
    size : int
        Sigma of a Gaussian window or cutoff frequency of a Butterworth window.
    window_name : str, optional
        Two options: "gaussian" or "butter".

    Returns
    -------
    array_like
        2D array of the window.
    """
    xcenter = np.ceil(width / 2.0) - 1.0
    xlist = np.arange(width) - xcenter
    if window_name == "butter":
        window = 1.0 - 1.0 / (1.0 + np.power(xlist / size, 2))
    else:
        window = 1.0 - np.exp(-xlist ** 2 / (2 * (size ** 2)))
    return np.tile(window, (height, 1))


def apply_wavelet_decomposition(mat, wavelet_name, level=None):
    """
    Apply 2D wavelet decomposition.

    Parameters
    ----------
    mat : array_like
        2D array.
    wavelet_name : str
        Name of a wavelet. E.g. "db5"
    level : int, optional
        Decomposition level. It is contrained to return an array with
        a minimum size of larger than 16 pixels.

    Returns
    -------
    list
        The first element is an 2D-array, next elements are tuples of three
        2D-arrays. i.e [mat_n, (cH_level_n, cV_level_n, cD_level_n), ...,
        (cH_level_1, cV_level_1, cD_level_1)]
    """
    (nrow, ncol) = mat.shape
    max_level = int(
        min(np.floor(np.log2(nrow / 16.0)), np.floor(np.log2(ncol / 16.0))))
    if (level is None) or (level > max_level) or (level < 1):
        level = max_level
    return pywt.wavedec2(mat, wavelet_name, level=level)


def apply_wavelet_recontruction(data, wavelet_name, ignore_level=None):
    """
    Apply 2D wavelet reconstruction.

    Parameters
    ----------
    data : list or tuple
        The first element is an 2D-array, next elements are tuples of three
        2D-arrays. i.e [mat_n, (cH_level_n, cV_level_n, cD_level_n), ...,
        (cH_level_1, cV_level_1, cD_level_1)].
    wavelet_name : str
        Name of a wavelet. E.g. "db5"
    ignore_level : int, optional
        Decomposition level to be ignored for reconstruction.

    Returns
    -------
    array_like
        2D array. Note that the sizes of the array are always even numbers.
    """
    if ignore_level is not None:
        level = len(data[1:])
        if level >= ignore_level > 0:
            data[-ignore_level] = tuple(
                [np.zeros_like(v) for v in data[-ignore_level]])
    return pywt.waverec2(data, wavelet_name)


def check_level(level, n_level):
    """
    Supplementary method for the method of "apply_filter_to_wavelet_component".
    To check if the provided level is in the right format.
    """
    if level is None:
        level = list(range(1, n_level + 1))
    else:
        if isinstance(level, int):
            if 0 < level <= n_level:
                level = [level]
            else:
                raise ValueError(
                    "Level is out of range: [1:{}]!".format(n_level))
        elif isinstance(level, list):
            level = [(i if (0 < i <= n_level) else 0) for i in level]
            if 0 in level:
                raise ValueError(
                    "Level is out of range: [1:{}]!".format(n_level))
        else:
            raise ValueError("Level-input is incorrect!")
    return level


def apply_filter_to_wavelet_component(data, level=None, order=1,
                                      method="gaussian_filter", para=[(1, 11)]):
    """
    Apply a filter to a component of the wavelet decomposition of an image.

    Parameters
    ----------
    data : list or tuple
        The first element is an 2D-array, next elements are tuples of three
        2D-arrays. i.e [mat_n, (cH_level_n, cV_level_n, cD_level_n), ...,
        (cH_level_1, cV_level_1, cD_level_1)].
    level : int, list of int, or None
        Decomposition level to be applied the filter.
    order : {0, 1, 2}
        Specify which component in a tuple, (cH_level_n, cV_level_n, cD_level_n)
        to be filtered.
    method : str
        Name of the filter in the namespace.
    para : list or tuple
        Parameters of the filter.

    Returns
    -------
    list or tuple
        The first element is an 2D-array, next elements are tuples of three
        2D-arrays. i.e [mat_n, (cH_level_n, cV_level_n, cD_level_n), ...,
        (cH_level_1, cV_level_1, cD_level_1)].
    """
    n_level = len(data[1:])
    level = check_level(level, n_level)
    order = np.clip(order, 0, 2)
    data = [list(i_data) for i_data in data]
    if not isinstance(para, list):
        para = tuple(list([para]))
    else:
        para = tuple(para)
    for i in level:
        if method in dir(ndi):
            data[i][order] = getattr(ndi, method)(data[i][order], *para)
        else:
            if method in dir():
                obj = sys.modules[__name__]
                data[i][order] = getattr(obj, method)(data[i][order], *para)
            else:
                raise ValueError("Can't find the method: '{}' in the"
                                 " namespace!".format(method))
    data = [tuple(i_data) for i_data in data]
    data[0] = np.asarray(data[0])
    return data


def interpolate_inside_stripe(mat, list_mask, kind="linear"):
    """
    Interpolate grayscales inside vertical stripes of an image. Stripe locations
    given by a binary 1D-mask.

    Parameters
    ----------
    mat : array_like
        2D array.
    list_mask : array_like
        1D array. Must equal the width of an image.
    kind : {'linear', 'cubic', 'quintic'}, optional
        The kind of spline interpolation to use. Default is 'linear'.

    Returns
    -------
    array_like
    """
    (nrow, ncol) = mat.shape
    if len(list_mask) != ncol:
        raise ValueError("Length of a binary 1D-mask is not the same as the "
                         "width of an image!")
    mat = np.copy(mat)
    list_mask = np.copy(list_mask)
    list_mask[0:2] = 0.0
    list_mask[-2:] = 0.0
    xlist = np.where(list_mask < 1.0)[0]
    ylist = np.arange(nrow)
    zmat = mat[:, xlist]
    finter = interpolate.interp2d(xlist, ylist, zmat, kind=kind)
    xlist_miss = np.where(list_mask > 0.0)[0]
    if len(xlist_miss) > 0:
        mat[:, xlist_miss] = finter(xlist_miss, ylist)
    return mat


def rectangular_from_polar(width_reg, height_reg, width_pol, height_pol):
    """
    Generate coordinates of a rectangular grid from polar coordinates.

    Parameters
    -----------
    width_reg : int
        Width of an image in the Cartesian coordinate system.
    height_reg : int
        Height of an image in the Cartesian coordinate system.
    width_pol : int
        Width of an image in the polar coordinate system.
    height_pol : int
        Height of an image in the polar coordinate system.

    Returns
    ------
    x_mat : array_like
         2D array. Broadcast of the x-coordinates.
    y_mat : array_like
         2D array. Broadcast of the y-coordinates.
    """
    xcenter = (width_reg - 1.0) * 0.5
    ycenter = (height_reg - 1.0) * 0.5
    r_max = np.floor(max(xcenter, ycenter))
    r_list = np.linspace(0.0, r_max, width_pol)
    theta_list = np.arange(0.0, height_pol, 1.0) * 2 * np.pi / (height_pol - 1)
    r_mat, theta_mat = np.meshgrid(r_list, theta_list)
    x_mat = np.float32(
        np.clip(xcenter + r_mat * np.cos(theta_mat), 0, width_reg - 1))
    y_mat = np.float32(
        np.clip(ycenter + r_mat * np.sin(theta_mat), 0, height_reg - 1))
    return x_mat, y_mat


def polar_from_rectangular(width_pol, height_pol, width_reg, height_reg):
    """
    Generate polar coordinates from grid coordinates.

    Parameters
    -----------
    width_pol : int
        Width of an image in the polar coordinate system.
    height_pol : int
        Height of an image in the polar coordinate system.
    width_reg : int
        Width of an image in the Cartesian coordinate system.
    height_reg : int
        Height of an image in the Cartesian coordinate system.

    Returns
    ------
    r_mat : array_like
         2D array. Broadcast of the r-coordinates.
    theta_mat : array_like
         2D array. Broadcast of the theta-coordinates.
    """
    xcenter = (width_reg - 1.0) * 0.5
    ycenter = (height_reg - 1.0) * 0.5
    r_max = np.floor(max(xcenter, ycenter))
    x_list = (np.flipud(np.arange(width_reg)) - xcenter) * width_pol / r_max
    y_list = (np.flipud(np.arange(height_reg)) - ycenter) * width_pol / r_max
    x_mat, y_mat = np.meshgrid(x_list, y_list)
    r_mat = np.float32(
        np.clip(np.sqrt(x_mat ** 2 + y_mat ** 2), 0, width_pol - 1))
    theta_mat = np.float32(np.clip(
        (np.pi + np.arctan2(y_mat, x_mat)) * (height_pol - 1) / (2 * np.pi), 0,
        height_pol - 1))
    return r_mat, theta_mat


def transform_slice_forward(mat, coord_mat=None):
    """
    Transform a reconstructed image into polar coordinates.

    Parameters
    ----------
    mat : array_like
        Square array. Reconstructed image.
    coord_mat : tuple of array_like, optional
        (Square array of x-coordinates , square array of y-coordinates) or
        generated if None.

    Returns
    -------
    array_like
        Transformed image.
    """
    (nrow, ncol) = mat.shape
    if nrow != ncol:
        raise ValueError("Height and width of the image is not the same!")
    if coord_mat is None:
        (x_mat, y_mat) = rectangular_from_polar(ncol, ncol, ncol, ncol)
    else:
        (x_mat, y_mat) = coord_mat
        if (x_mat.shape != mat.shape) or (y_mat.shape != mat.shape):
            raise ValueError("Shape of the coordinate array is not the same as"
                             " the shape of the image!")
    return mapping(mat, x_mat, y_mat)


def transform_slice_backward(mat, coord_mat=None):
    """
    Transform a reconstructed image in polar coordinates back to rectangular
    coordinates.

    Parameters
    ----------
    mat : array_like
        Square array. Reconstructed image in polar coordinates.
    coord_mat : tuple of array_like, optional
        (Square array of r-coordinates , square array of theta-coordinates) or
        generated if None.

    Returns
    -------
    array_like
        Transformed image.
    """
    (nrow, ncol) = mat.shape
    if nrow != ncol:
        raise ValueError("Height and width of the image is not the same!")
    if coord_mat is None:
        (r_mat, theta_mat) = polar_from_rectangular(ncol, ncol, ncol, ncol)
    else:
        (r_mat, theta_mat) = coord_mat
        if (r_mat.shape != mat.shape) or (theta_mat.shape != mat.shape):
            raise ValueError("Shape of the coordinate array is not the same as"
                             " the shape of the image!")
    return mapping(mat, r_mat, theta_mat)


def make_2d_gaussian_window(height, width, sigmax, sigmay):
    """
    Create a 2D Gaussian window.

    Parameters
    ----------
    height : int
        Height of the image.
    width : int
        Width of the image.
    sigmax : int
        Sigma in the x-direction.
    sigmay : int
        Sigma in the y-direction.

    Returns
    -------
    array_like
        2D array.
    """
    xcenter = (width - 1.0) / 2.0
    ycenter = (height - 1.0) / 2.0
    y, x = np.ogrid[-ycenter:height - ycenter, -xcenter:width - xcenter]
    window = np.exp(-(x ** 2 / (2 * sigmax ** 2) + y ** 2 / (2 * sigmay ** 2)))
    return window


def apply_gaussian_filter(mat, sigmax, sigmay, pad=None, mode=None):
    """
    Filtering an image in the Fourier domain using a 2D Gaussian window.
    Smaller is stronger.

    Parameters
    ----------
    mat : array_like
        2D array.
    sigmax : int
        Sigma in the x-direction.
    sigmay : int
        Sigma in the y-direction.
    pad : int or None
        Padding for the Fourier transform.
    mode : str, list of str, or tuple of str
        Padding method. One of options : 'reflect', 'edge', 'constant'. Full
        list is at:
        https://numpy.org/doc/stable/reference/generated/numpy.pad.html

    Returns
    -------
    array_like
        2D array. Filtered image.
    """
    if mode is None:
        # Default for a sinogram image.
        mode1 = "edge"
        mode2 = "mean"
    else:
        if isinstance(mode, list) or isinstance(mode, tuple):
            mode1 = mode[0]
            mode2 = mode[1]
        else:
            mode1 = mode2 = mode
    if pad is None:
        pad = min(150, int(0.1 * min(mat.shape)))
    mat_pad = np.pad(mat, ((0, 0), (pad, pad)), mode=mode1)
    mat_pad = np.pad(mat_pad, ((pad, pad), (0, 0)), mode=mode2)
    (nrow, ncol) = mat_pad.shape
    window = make_2d_gaussian_window(nrow, ncol, sigmax, sigmay)
    listx = np.arange(0, ncol)
    listy = np.arange(0, nrow)
    x, y = np.meshgrid(listx, listy)
    mat_sign = np.power(-1.0, x + y)
    mat_filt = np.real(
        fft.ifft2(fft.fft2(mat_pad * mat_sign) * window) * mat_sign)
    return mat_filt[pad:nrow - pad, pad:ncol - pad]


def apply_1d_regularizer(list_data, sijmat):
    """
    Supplementary method for the method of "apply_regularization_filter".
    To apply a regularizer to an 1D-array.
    """
    ncol = len(list_data)
    list_grad = np.zeros(ncol, dtype=np.float32)
    list_grad[1:-1] = (-1) * np.diff(list_data, 2)
    list_grad[0] = list_data[0] - list_data[1]
    list_grad[-1] = list_data[-1] - list_data[-2]
    list_coe = np.sum(np.tile(list_grad, (ncol, 1)) * sijmat, axis=1)
    return list_data + list_coe


def apply_regularization_filter(mat, alpha, axis=1, ncore=None):
    """
    Apply a regularization filter using the method in Ref. [1].
    Note that it's computationally costly.

    Parameters
    ----------
    mat : array_like
        2D array
    alpha : float
        Regularization parameter, e.g. 0.001. Smaller is stronger.
    axis : int
        Axis along which to apply the filter.
    ncore: int or None
        Number of cores used for computing. Automatically selected if None.

    Returns
    -------
    array_like
        2D array. Smoothed image.

    References
    ----------
    .. [1] https://doi.org/10.1016/j.aml.2010.08.022
    """
    if ncore is None:
        ncore = np.clip(mp.cpu_count() - 1, 1, None)
    if axis == 0:
        mat = np.transpose(mat)
    (nrow, ncol) = mat.shape
    sijmat = calculate_regularization_coefficient(ncol, alpha)
    mat = np.asarray(Parallel(n_jobs=ncore, backend="threading")(
        delayed(apply_1d_regularizer)(mat[i], sijmat) for i in range(nrow)))
    if axis == 0:
        mat = np.transpose(mat)
    return mat


def transform_1d_window_to_2d(win_1d):
    """
    Transform a 1d-window to 2d-window.
    Useful for designing a Fourier filter.

    Parameters
    ----------
    win_1d : array_like
        1D array.

    Returns
    --------
    win_2d : array_like
        Square array, a 2D version of the 1d-window.
    """
    width0 = len(win_1d)
    if width0 % 2 == 0:
        width = width0 + 1
    else:
        width = width0
    center = width // 2
    x_list = (1.0 * np.flipud(np.arange(width)) - center)
    y_list = (1.0 * np.arange(width) - center)
    x_mat, y_mat = np.meshgrid(x_list, y_list)
    r_mat = np.float32(np.clip(np.sqrt(x_mat ** 2 + y_mat ** 2), 0, center))
    theta_mat = np.arctan2(y_mat, x_mat)
    r_mat[theta_mat < 0] *= -1
    r_mat = np.float32(np.clip(r_mat + center, 0, width - 1))
    theta_mat = np.clip(np.float32(theta_mat * (width - 1.0) / np.pi), 0,
                        width - 1)
    mat = np.tile(win_1d, (width, 1))
    win_2d = mapping(mat, r_mat, theta_mat)
    return win_2d[0:width0, 0:width0]


def detect_sample(sinogram, sino_type="180"):
    """
    To check if there is a sample in a sinogram using the "double-wedge"
    property of the Fourier transform of the sinogram (Ref. [1]).

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    sino_type : {"180", "360"}
        Sinogram type : 180-degree or 360-degree.

    Returns
    -------
    bool
        True if there is a sample.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448

    """
    check = True
    if not (sino_type == "180" or sino_type == "360"):
        raise ValueError("!!! Use only one of two options: '180' or '360'!!!")
    if sino_type == "180":
        sinogram = 1.0 * np.vstack((sinogram, np.fliplr(sinogram)))
    sino_fft = np.abs(fft.fftshift(fft.fft2(sinogram)))
    (nrow, ncol) = sino_fft.shape
    ycenter = nrow // 2
    xcenter = ncol // 2
    radi = min(20, int(np.ceil(0.05 * min(ycenter, xcenter))))
    sino_fft = sino_fft[:ycenter - radi, :xcenter - radi]
    size = min(30, int(np.ceil(0.02 * min(nrow, ncol))))
    sino_smooth = ndi.gaussian_filter(sino_fft, size)
    (nrow1, _) = sino_smooth.shape
    row = int(0.8 * nrow1)
    list_check = np.zeros(nrow1 - row, dtype=np.float32)
    pos = np.argmax(sino_smooth[row])
    for i in np.arange(row, nrow1):
        pos1 = np.argmax(sino_smooth[i])
        if pos1 > pos:
            list_check[i - row] = 1.0
    ratio = (np.sum(list_check) / len(list_check))
    if ratio < 0.4:
        check = False
    return check


def fix_non_sample_areas(overlap_metadata):
    """
    Used to fix overlap values of grid-cells without sample by copying from
    its neighbours

    Parameters
    ---------
    overlap_metadata : array_like
        A matrix of overlap values of each grid-cell where each element is a
        list of [overlap, side].

    Returns
    -------
    metadata : array_like
    """
    (g_nrow, g_ncol, _) = overlap_metadata.shape
    metadata = np.copy(overlap_metadata)
    for i in np.arange(g_nrow):
        i1 = i - 1
        i2 = i + 1
        for j in np.arange(g_ncol):
            (area, _) = overlap_metadata[i, j]
            j1 = j - 1
            j2 = j + 1
            if area == 0:
                if 0 <= i1 < g_nrow:
                    (area1, side1) = overlap_metadata[i1, j]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
                if 0 <= i2 < g_nrow:
                    (area1, side1) = overlap_metadata[i2, j]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
                if 0 <= j1 < g_ncol:
                    (area1, side1) = overlap_metadata[i, j1]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
                if 0 <= j2 < g_ncol:
                    (area1, side1) = overlap_metadata[i, j2]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
    # Run the same above routine but in reverse order.
    for i in np.arange(g_nrow - 1, -1, -1):
        i1 = i - 1
        i2 = i + 1
        for j in np.arange(g_ncol - 1, -1, -1):
            (area, _) = metadata[i, j]
            j1 = j - 1
            j2 = j + 1
            if area == 0:
                if 0 <= i1 < g_nrow:
                    (area1, side1) = metadata[i1, j]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
                if 0 <= i2 < g_nrow:
                    (area1, side1) = metadata[i2, j]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
                if 0 <= j1 < g_ncol:
                    (area1, side1) = metadata[i, j1]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
                if 0 <= j2 < g_ncol:
                    (area1, side1) = metadata[i, j2]
                    if area1 != 0:
                        metadata[i, j] = np.asarray([area1, side1])
                        continue
        return metadata


def locate_slice(slice_idx, height, overlap_metadata):
    """
    Locate slice indices in grid-rows given a slice index of the reconstruction
    data as a whole.

    Parameters
    ----------
    slice_idx : int
        Slice index of full reconstruction data.
    height : int
        Height of a projection image of each grid-cell.
    overlap_metadata : array_like
        A matrix of overlap values of each grid-row where each element is a
        list of [overlap, side]. Used to stitch the grid-data along the
        row-direction.

    Returns
    -------
    list of int and float
        If the slice is not in the overlapping area between two grid-rows, the
        result is a list of [grid_row_index, slice_index, weight_factor]. If the
        slice is in the overlapping area between two grid-rows, the result is a
        list of [[grid_row_index_0, slice_index_0, weight_factor_0],
        [grid_row_index_1, slice_index_1, weight_factor_1]]
    """
    g_nrow = overlap_metadata.shape[0] + 1
    side = overlap_metadata[0, 0, 1]
    overlap_list = overlap_metadata[:, 0, 0]
    if side == 1:
        list_slices = [(np.arange(i * height, i * height + height) -
                        np.sum(overlap_list[0: i])) for i in range(g_nrow)]
    else:
        list_slices = [(np.arange(i * height + height - 1, i * height - 1, -1) -
                        np.sum(overlap_list[0: i])) for i in range(g_nrow)]
    list_slices = np.asarray(list_slices)
    results = []
    for i, list1 in enumerate(list_slices):
        pos = np.squeeze(np.where(list1 == slice_idx)[0])
        if pos.size == 1:
            results.append([i, pos, 1.0])
    if len(results) == 2:
        if side == 1:
            results[0][2] = (1.0 * list_slices[results[0][0]][
                -1] - slice_idx) / (overlap_list[results[0][0]] - 1.0)
            results[1][2] = (slice_idx - 1.0 * list_slices[results[1][0]][
                0]) / (overlap_list[results[0][0]] - 1.0)
        else:
            results[0][2] = (- slice_idx + 1.0 * list_slices[results[0][0]][
                0]) / (overlap_list[results[0][0]] - 1.0)
            results[1][2] = (-1.0 * list_slices[results[1][0]][
                -1] + slice_idx) / (overlap_list[results[0][0]] - 1.0)
    return results


def locate_slice_chunk(slice_start, slice_stop, height, overlap_metadata):
    """
    Locate slice indices in grid-rows given slice indices of the reconstruction
    data as a whole.

    Parameters
    ----------
    slice_start : int
        Starting index of full reconstruction data.
    slice_stop : int
        Stopping index of full reconstruction data.
    height : int
        Height of a projection image of each grid-cell.
    overlap_metadata : array_like
        A matrix of overlap values of each grid-row where each element is a
        list of [overlap, side]. Used to stitch the grid-data along the
        row-direction.

    Returns
    -------
    list of list of int and float
        List of results for each slice index. If a slice is not in the
        overlapping area between two grid-rows, the result is a list of
        [grid_row_index, slice_index, weight_factor]. If a slice is in the
        overlapping area between two grid-rows, the result is a list of
        [[grid_row_index_0, slice_index_0, weight_factor_0],
        [grid_row_index_1, slice_index_1, weight_factor_1]].
    """
    if slice_stop < slice_start:
        raise ValueError(
            "Stopping index must be larger than the starting index!!!")
    g_nrow = overlap_metadata.shape[0] + 1
    side = overlap_metadata[0, 0, 1]
    overlap_list = overlap_metadata[:, 0, 0]
    if side == 1:
        list_slices = [(np.arange(i * height, i * height + height) -
                        np.sum(overlap_list[0: i])) for i in range(g_nrow)]
    else:
        list_slices = [(np.arange(i * height + height - 1, i * height - 1, -1) -
                        np.sum(overlap_list[0: i])) for i in range(g_nrow)]
    list_slices = np.asarray(list_slices)
    results = []
    for i, list1 in enumerate(list_slices):
        result1 = []
        if side == 1:
            for slice_idx in range(slice_start, slice_stop):
                pos = np.squeeze(np.where(list1 == slice_idx)[0])
                if pos.size == 1:
                    fact = 1.0
                    if i == 0:
                        ver_overlap = overlap_list[i]
                        dis1 = len(list1) - pos - 1
                        if dis1 < ver_overlap:
                            fact = dis1 / (ver_overlap - 1)
                    elif i == (g_nrow - 1):
                        ver_overlap = overlap_list[i - 1]
                        if pos < ver_overlap:
                            fact = pos / (ver_overlap - 1)
                    else:
                        ver_overlap1 = overlap_list[i]
                        dis1 = len(list1) - pos - 1
                        if dis1 < ver_overlap1:
                            fact = dis1 / (ver_overlap1 - 1)
                        if pos < ver_overlap1:
                            fact = pos / (ver_overlap1 - 1)
                        ver_overlap2 = overlap_list[i - 1]
                        dis1 = len(list1) - pos - 1
                        if dis1 < ver_overlap2:
                            fact = dis1 / (ver_overlap2 - 1)
                        if pos < ver_overlap2:
                            fact = pos / (ver_overlap2 - 1)
                    result1.append([i, pos, fact])
        else:
            for slice_idx in range(slice_start, slice_stop):
                pos = np.squeeze(np.where(list1 == slice_idx)[0])
                if pos.size == 1:
                    fact = 1.0
                    if i == 0:
                        ver_overlap = overlap_list[i]
                        if pos < ver_overlap:
                            fact = 1.0 * pos / (ver_overlap - 1)
                    elif i == (g_nrow - 1):
                        ver_overlap = overlap_list[i - 1]
                        dis1 = len(list1) - pos - 1
                        if dis1 < ver_overlap:
                            fact = 1.0 * dis1 / (ver_overlap - 1)
                    else:
                        ver_overlap1 = overlap_list[i]
                        dis1 = len(list1) - pos - 1
                        if dis1 < ver_overlap1:
                            fact = 1.0 * dis1 / (ver_overlap1 - 1)
                        if pos < ver_overlap1:
                            fact = 1.0 * pos / (ver_overlap1 - 1)
                        ver_overlap2 = overlap_list[i - 1]
                        dis1 = len(list1) - pos - 1
                        if dis1 < ver_overlap2:
                            fact = 1.0 * dis1 / (ver_overlap2 - 1)
                        if pos < ver_overlap2:
                            fact = 1.0 * pos / (ver_overlap2 - 1)
                    result1.append([i, pos, fact])
        if len(result1) > 0:
            results.append(result1)
    return results
